#!/bin/python
import roomai.common

AllSevenKingPatterns = dict()
###
###numCards
AllSevenKingPatterns["p_0"] = ("p_0", 0) ## check
AllSevenKingPatterns["p_1"] = ("p_1", 1)
AllSevenKingPatterns["p_2"] = ("p_2", 2)
AllSevenKingPatterns["p_3"] = ("p_3", 3)
AllSevenKingPatterns["p_4"] = ("p_4", 4)


point_str_to_rank  = {'7':14, 'R':13, 'r':12, '5':11,  '2':10,  '3':9,  'A':8,  'K':7,\
                      'Q':6,  'J':5,   'T':4,   '9':3,   '8':2,   '6':1,   '4':0}
point_rank_to_str  = {14:'7', 13:'R', 12:'r',  11:'5', 10:'2',  9:'3',  8:'A',   7:'K',\
                       6:'Q',  5:'J',   4:'T',   3:'9',   2:'8',   1:'6',  0:'4'}
suit_str_to_rank   = {'Spade':3, 'Heart':2, 'Diamond':1, 'Club':0,  'ForKing':4}
suit_rank_to_str   = {3:'Spade', 2: 'Heart', 1: 'Diamond', 0:'Club', 4:'ForKing'}

class SevenKingPokerCard(roomai.common.PokerCard):
    """
    """
    @property
    def point_rank(self):
        """

        Returns:

        """
        return point_str_to_rank[self.point_str]
    @property
    def suit_rank(self):
        """

        Returns:

        """
        return suit_str_to_rank[self.suit_str]
    def lookup(cls, key):
        """

        Args:
            key:

        Returns:

        """
        return AllSevenKingPokerCards[key]
    def __deepcopy__(self,  memodict={}, newinstance = None):
        """

        Args:
            memodict:
            newinstance:

        Returns:

        """
        if newinstance is None:
            newinstance = AllSevenKingPokerCards[self.key]
        else:
            newinstance = super(SevenKingPokerCard, self).__deepcopy__(newinstance = newinstance)
        return newinstance

AllSevenKingPokerCards = dict()
for point in point_str_to_rank:
    if point != "r" and point != "R":
        for suit in suit_str_to_rank:
            if suit != "ForKing":
                AllSevenKingPokerCards["%s_%s" % (point, suit)] = SevenKingPokerCard("%s_%s" % (point, suit))
AllSevenKingPokerCards["R_ForKing"] = SevenKingPokerCard("R_ForKing")
AllSevenKingPokerCards["r_ForKing"] = SevenKingPokerCard("r_ForKing")



#!/bin/python
import roomai
import random
from roomai.doudizhu.DouDiZhuPokerUtils import *

class DouDiZhuPokerRandomPlayer(roomai.common.AbstractPlayer):
    """
    """
    
    def __init__(self):
        """

        """
        self.available_actions  = None
    
    #@override    
    def receive_info(self,info):
        """

        Args:
            info:
        """
        self.available_actions = info.person_state.available_actions


    #@override
    def take_action(self):
        """

        Returns:

        """
        candidates = self.available_actions.values()
        idx = int(random.random() * len(candidates))
        action = candidates[idx]
        return action


    #@override
    def reset(self, action):
        """

        Args:
            action:
        """
        pass



#!/bin/python
#coding:utf-8

import random
import copy
import roomai.common
import roomai
import logging

from roomai.common import Info
from TexasHoldemUtil import *


class TexasHoldemEnv(roomai.common.AbstractEnv):
    """
    """

    @classmethod
    def check_initialization_configuration(cls, env):
        """

        Args:
            env:

        Returns:

        """
        if len(env.chips) != env.num_players:
            raise ValueError("len(env.chips)%d != env.num_players%d" % (len(env.chips), env.num_players))

        if env.num_players * 7 > 52:
            raise ValueError("env.num_players * 5 must be less than 51, now env.num_players = %d" % (env.num_players))

        return True

    #@override
    def init(self, params = dict()):
        """

        Args:
            params:

        Returns:

        """
        self.logger         = roomai.get_logger()

        if "num_players" in params:
            self.num_players = params["num_players"]
        else:
            self.num_players = 3

        if "dealer_id" in params:
            self.dealer_id = params["dealer_id"]
        else:
            self.dealer_id = int(random.random() * self.num_players)

        if "chips" in params:
            self.chips     = params["chips"]
        else:
            self.chips     = [1000 for i in range(self.num_players)]

        if "big_blind_bet" in params:
            self.big_blind_bet = params["big_blind_bet"]
        else:
            self.big_blind_bet = 10

        if "allcards" in params:
            self.allcards = [c.__deepcopy__() for c in params["allcards"]]
        else:
            self.allcards = []
            for i in xrange(13):
                for j in xrange(4):
                    self.allcards.append(roomai.common.PokerCard(i, j))
            random.shuffle(self.allcards)

        if "record_history" in params:
            self.record_history = params["record_history"]
        else:
            self.record_history = False


        self.check_initialization_configuration(self)

        hand_cards       = []
        for i in xrange(self.num_players):
              hand_cards.append(self.allcards[i*2:(i+1)*2])
        keep_cards   = self.allcards[self.num_players*2:self.num_players*2+5]

        ## public info
        small = (self.dealer_id + 1) % self.num_players
        big   = (self.dealer_id + 2) % self.num_players

        self.public_state                       = TexasHoldemPublicState()
        self.public_state.num_players           = self.num_players
        self.public_state.dealer_id             = self.dealer_id
        self.public_state.big_blind_bet         = self.big_blind_bet
        self.public_state.raise_account         = self.big_blind_bet

        self.public_state.is_fold               = [False for i in xrange(self.num_players)]
        self.public_state.num_quit              = 0
        self.public_state.is_allin              = [False for i in xrange(self.num_players)]
        self.public_state.num_allin             = 0
        self.public_state.is_needed_to_action   = [True for i in xrange(self.num_players)]
        self.public_state.num_needed_to_action  = self.public_state.num_players

        self.public_state.bets                  = [0 for i in xrange(self.num_players)]
        self.public_state.chips                 = self.chips
        self.public_state.stage                 = StageSpace.firstStage
        self.public_state.turn                  = (big+1)%self.public_state.num_players
        self.public_state.public_cards          = []

        self.public_state.previous_id           = None
        self.public_state.previous_action       = None

        if self.public_state.chips[big] > self.big_blind_bet:
            self.public_state.chips[big] -= self.big_blind_bet
            self.public_state.bets[big]  += self.big_blind_bet
        else:
            self.public_state.bets[big]     = self.public_state.chips[big]
            self.public_state.chips[big]    = 0
            self.public_state.is_allin[big] = True
            self.public_state.num_allin    += 1
        self.public_state.max_bet_sofar     = self.public_state.bets[big]
        self.public_state.raise_account = self.big_blind_bet

        if self.public_state.chips[small] > self.big_blind_bet / 2:
            self.public_state.chips[small] -= self.big_blind_bet /2
            self.public_state.bets[small]  += self.big_blind_bet /2
        else:
            self.public_state.bets[small]     = self.public_state.chips[small]
            self.public_state.chips[small]    = 0
            self.public_state.is_allin[small] = True
            self.public_state.num_allin      += 1

        self.public_state.is_terminal         = False
        self.public_state.scores              = None

        # private info
        self.private_state = TexasHoldemPrivateState()
        self.private_state.hand_cards       = [[] for i in xrange(self.num_players)]
        for i in xrange(self.num_players):
            self.private_state.hand_cards[i]  = [hand_cards[i][j].__deepcopy__() for j in xrange(len(hand_cards[i]))]
        self.private_state.keep_cards         = [keep_cards[i].__deepcopy__() for i in xrange(len(keep_cards))]

        ## person info
        self.person_states                      = [TexasHoldemPersonState() for i in xrange(self.num_players)]
        for i in xrange(self.num_players):
            self.person_states[i].id = i
            self.person_states[i].hand_cards = [hand_cards[i][j].__deepcopy__() for j in xrange(len(hand_cards[i]))]
        self.person_states[self.public_state.turn].available_actions = self.available_actions(self.public_state, self.person_states[self.public_state.turn])

        self.__gen_history__()
        infos = self.__gen_infos__()

        if self.logger.level <= logging.DEBUG:
            self.logger.debug("TexasHoldemEnv.init: num_players = %d, dealer_id = %d, chip = %d, big_blind_bet = %d"%(\
                self.public_state.num_players,\
                self.public_state.dealer_id,\
                self.public_state.chips[0],\
                self.public_state.big_blind_bet
            ))

        return infos, self.public_state, self.person_states, self.private_state

    ## we need ensure the action is valid
    #@Overide
    def forward(self, action):
        """
        Args:
            action: 

        Raises:
            None: throw ValueError when the action is invalid at this time
        """
        pu         = self.public_state
        pe         = self.person_states
        pr         = self.private_state

        if not self.is_action_valid(action, pu, pe[pu.turn]):
            self.logger.critical("action=%s is invalid" % (action.key))
            raise ValueError("action=%s is invalid" % (action.key))


        if action.option == TexasHoldemAction.Fold:
            self.action_fold(action)
        elif action.option == TexasHoldemAction.Check:
            self.action_check(action)
        elif action.option == TexasHoldemAction.Call:
            self.action_call(action)
        elif action.option == TexasHoldemAction.Raise:
            self.action_raise(action)
        elif action.option == TexasHoldemAction.AllIn:
            self.action_allin(action)
        else:
            raise Exception("action.option(%s) not in [Fold, Check, Call, Raise, AllIn]"%(action.option))
        pu.previous_id     = pu.turn
        pu.previous_action = action
        pu.is_terminal     = False
        pu.scores          = None

        # computing_score
        if TexasHoldemEnv.is_compute_scores(self.public_state):
            ## need showdown
            pu.public_cards = pr.keep_cards[0:5]
            pu.is_terminal  = True
            pu.scores       = self.compute_scores()


        # enter into the next stage
        elif TexasHoldemEnv.is_nextround(self.public_state):
            add_cards = []
            if pu.stage == StageSpace.firstStage:   add_cards = pr.keep_cards[0:3]
            if pu.stage == StageSpace.secondStage:  add_cards = [pr.keep_cards[3]]
            if pu.stage == StageSpace.thirdStage:   add_cards = [pr.keep_cards[4]]

            pu.public_cards.extend(add_cards)
            pu.stage                      = pu.stage + 1

            pu.num_needed_to_action       = 0
            pu.is_needed_to_action        = [False for i in xrange(pu.num_players)]
            for i in xrange(pu.num_players):
                if pu.is_fold[i] != True and pu.is_allin[i] != True:
                    pu.is_needed_to_action[i]      = True
                    pu.num_needed_to_action       += 1

            pu.turn                                             = pu.dealer_id
            pu.turn                                             = self.next_player(pu)
            pe[self.public_state.previous_id].available_actions = dict()
            pe[self.public_state.turn].available_actions        = self.available_actions(self.public_state, self.person_states[self.public_state.turn])

        ##normal
        else:
            pu.turn                                                             = self.next_player(pu)
            self.person_states[self.public_state.previous_id].available_actions = dict()
            self.person_states[self.public_state.turn].available_actions        = self.available_actions(self.public_state, self.person_states[self.public_state.turn])


        if self.logger.level <= logging.DEBUG:
            self.logger.debug("TexasHoldemEnv.forward: num_quit+num_allin = %d+%d = %d, action = %s, stage = %d"%(\
                self.public_state.num_quit,\
                self.public_state.num_allin,\
                self.public_state.num_quit + self.public_state.num_allin,\
                action.key,\
                self.public_state.stage\
            ))

        self.__gen_history__()
        infos = self.__gen_infos__()
        return infos, self.public_state, self.person_states, self.private_state

    #override
    @classmethod
    def compete(cls, env, players):
        """

        Args:
            env:
            players:

        Returns:

        """

        total_scores       = [0    for i in xrange(len(players))]
        total_count        = 1000


        for count in xrange(total_count):

            chips          = [(1000 + int(random.random() * 200)) for i in range(len(players))]
            num_players    = len(players)
            dealer_id      = int(random.random() * len(players))
            big_blind_bet  = 50

            infos, public, persons, private = env.init({"chips":chips,
                                                        "num_players":num_players,
                                                        "dealer_id":dealer_id,
                                                        "big_blind_bet":big_blind_bet})
            for i in xrange(len(players)):
                players[i].receive_info(infos[i])
            while public.is_terminal == False:
                turn = public.turn
                action = players[turn].take_action()
                #print len(infos[turn].person_state.available_actions),action.key(),turn
                infos, public, persons, private = env.forward(action)
                for i in xrange(len(players)):
                    players[i].receive_info(infos[i])

            for i in xrange(len(players)):
                players[i].receive_info(infos[i])
                total_scores[i] += public.scores[i]

            '''
            if count < 1000:
                print count,public.dealer_id,public.scores,public.stage
                for i in xrange(public.num_players):
                    for j in xrange(len(private.hand_cards[i])):
                        print private.hand_cards[i][j].key,
                    print ""
                print len(public.public_cards)
                for j in xrange(len(public.public_cards)):
                    print public.public_cards[j].key,
                print ""
                for i in xrange(public.num_players):
                    x = cls.cards2pattern(private.hand_cards[i], public.public_cards)
                    print x[0],x[5]
            '''

            if (count + 1)%500 == 0:
                tmp_scores = [0 for i in xrange(len(total_scores))]
                for i in xrange(len(total_scores)):
                    tmp_scores[i] = total_scores[i] / (count+1)
                roomai.get_logger().info("TexasHoldem completes %d competitions, scores=%s"%(count+1, ",".join([str(i) for i in tmp_scores])))

        for i in xrange(len(total_scores)):
            total_scores[i] /= 1.0 * total_count

        return total_scores


    def compute_scores(self):
        """
        Returns:
            a score array
        """
        pu = self.public_state
        pr = self.private_state

        ## compute score before showdown, the winner takes all
        if pu.num_players  ==  pu.num_quit + 1:
            scores = [0 for i in xrange(pu.num_players)]
            for i in xrange(pu.num_players):
                if pu.is_fold[i] == False:
                    scores[i] = sum(pu.bets)
                    break

        ## compute score after showdown
        else:
            scores                = [0 for i in xrange(pu.num_players)]
            playerid_pattern_bets = [] #for not_quit players
            for i in xrange(pu.num_players):
                if pu.is_fold[i] == True: continue
                hand_pattern = self.cards2pattern(pr.hand_cards[i], pr.keep_cards)
                playerid_pattern_bets.append((i,hand_pattern,pu.bets[i]))
            playerid_pattern_bets.sort(key=lambda x:x[1], cmp=self.compare_patterns)

            pot_line = 0
            previous = None
            tmp_playerid_pattern_bets      = []
            for i in xrange(len(playerid_pattern_bets)-1,-1,-1):
                if previous == None:
                    tmp_playerid_pattern_bets.append(playerid_pattern_bets[i])
                    previous = playerid_pattern_bets[i]
                elif self.compare_patterns(playerid_pattern_bets[i][1], previous[1]) == 0:
                    tmp_playerid_pattern_bets.append(playerid_pattern_bets[i])
                    previous = playerid_pattern_bets[i]
                else:
                    tmp_playerid_pattern_bets.sort(key = lambda x:x[2])
                    for k in xrange(len(tmp_playerid_pattern_bets)):
                        num1          = len(tmp_playerid_pattern_bets) - k
                        sum1          = 0
                        max_win_score = pu.bets[tmp_playerid_pattern_bets[k][0]]
                        for p in xrange(pu.num_players):
                            sum1      += min(max(0, pu.bets[p] - pot_line), max_win_score)
                        for p in xrange(k, len(tmp_playerid_pattern_bets)):
                            scores[tmp_playerid_pattern_bets[p][0]] += sum1 / num1
                        scores[pu.dealer_id] += sum1 % num1
                        if pot_line <= max_win_score:
                            pot_line = max_win_score
                    tmp_playerid_pattern_bets = []
                    tmp_playerid_pattern_bets.append(playerid_pattern_bets[i])
                    previous = playerid_pattern_bets[i]


            if len(tmp_playerid_pattern_bets) > 0:
                tmp_playerid_pattern_bets.sort(key = lambda  x:x[2])
                for i in xrange(len(tmp_playerid_pattern_bets)):
                    num1 = len(tmp_playerid_pattern_bets) - i
                    sum1 = 0
                    max_win_score = pu.bets[tmp_playerid_pattern_bets[i][0]]
                    for p in xrange(pu.num_players):
                        sum1 += min(max(0, pu.bets[p] - pot_line), max_win_score)
                    for p in xrange(i, len(tmp_playerid_pattern_bets)):
                        scores[tmp_playerid_pattern_bets[p][0]] += sum1 / num1
                    scores[pu.dealer_id] += sum1 % num1
                    if pot_line <= max_win_score: pot_line = max_win_score

        for p in xrange(pu.num_players):
            pu.chips[p] += scores[p]
            scores[p]   -= pu.bets[p]
        for p in xrange(pu.num_players):
            scores[p]   /= pu.big_blind_bet * 1.0
        return scores


    def action_fold(self, action):
        """

        Args:
            action:
        """
        pu = self.public_state
        pu.is_fold[pu.turn]             = True
        pu.num_quit                    += 1

        pu.is_needed_to_action[pu.turn] = False
        pu.num_needed_to_action        -= 1

    def action_check(self, action):
        """

        Args:
            action:
        """
        pu = self.public_state
        pu.is_needed_to_action[pu.turn] = False
        pu.num_needed_to_action        -= 1

    def action_call(self, action):
        """

        Args:
            action:
        """
        pu = self.public_state
        pu.chips[pu.turn] -= action.price
        pu.bets[pu.turn]  += action.price
        pu.is_needed_to_action[pu.turn] = False
        pu.num_needed_to_action        -= 1

    def action_raise(self, action):
        """

        Args:
            action:
        """
        pu = self.public_state


        pu.raise_account   = action.price + pu.bets[pu.turn] - pu.max_bet_sofar
        pu.chips[pu.turn] -= action.price
        pu.bets[pu.turn]  += action.price
        pu.max_bet_sofar   = pu.bets[pu.turn]

        pu.is_needed_to_action[pu.turn] = False
        pu.num_needed_to_action        -= 1
        p = (pu.turn + 1)%pu.num_players
        while p != pu.turn:
            if pu.is_allin[p] == False and pu.is_fold[p] == False and pu.is_needed_to_action[p] == False:
                pu.num_needed_to_action   += 1
                pu.is_needed_to_action[p]  = True
            p = (p + 1) % pu.num_players


    def action_allin(self, action):
        """

        Args:
            action:
        """
        pu = self.public_state

        pu.is_allin[pu.turn]   = True
        pu.num_allin          += 1

        pu.bets[pu.turn]      += action.price
        pu.chips[pu.turn]      = 0

        pu.is_needed_to_action[pu.turn] = False
        pu.num_needed_to_action        -= 1
        if pu.bets[pu.turn] > pu.max_bet_sofar:
            pu.max_bet_sofar = pu.bets[pu.turn]
            p = (pu.turn + 1) % pu.num_players
            while p != pu.turn:
                if pu.is_allin[p] == False and pu.is_fold[p] == False and pu.is_needed_to_action[p] == False:
                    pu.num_needed_to_action  += 1
                    pu.is_needed_to_action[p] = True
                p = (p + 1) % pu.num_players

            pu.max_bet_sofar = pu.bets[pu.turn]

#####################################Utils Function ##############################

    @classmethod
    def next_player(self, pu):
        """

        Args:
            pu:

        Returns:

        """
        i = pu.turn
        if pu.num_needed_to_action == 0:
            return -1

        p = (i+1)%pu.num_players
        while pu.is_needed_to_action[p] == False:
            p = (p+1)%pu.num_players
        return p

    @classmethod
    def is_compute_scores(self, pu):
        '''
        :return: 
        A boolean variable indicates whether is it time to compute scores
        '''

        if pu.num_players == pu.num_quit + 1:
            return True

        # below need showdown

        if pu.num_players <=  pu.num_quit + pu.num_allin +1 and pu.num_needed_to_action == 0:
            return True

        if pu.stage == StageSpace.fourthStage and self.is_nextround(pu):
            return True

        return False

    @classmethod
    def is_nextround(self, public_state):
        '''
        :return: 
        A boolean variable indicates whether is it time to enter the next stage
        '''
        return public_state.num_needed_to_action == 0

    @classmethod
    def cards2pattern(cls, hand_cards, remaining_cards):
        """

        Args:
            hand_cards:
            remaining_cards:

        Returns:

        """
        pointrank2cards = dict()
        for c in hand_cards + remaining_cards:
            if c.get_point_rank() in pointrank2cards:
                pointrank2cards[c.get_point_rank()].append(c)
            else:
                pointrank2cards[c.get_point_rank()] = [c]
        for p in pointrank2cards:
            pointrank2cards[p].sort(roomai.common.PokerCard.compare)

        suitrank2cards = dict()
        for c in hand_cards + remaining_cards:
            if c.get_suit_rank() in suitrank2cards:
                suitrank2cards[c.get_suit_rank()].append(c)
            else:
                suitrank2cards[c.get_suit_rank()] = [c]
        for s in suitrank2cards:
            suitrank2cards[s].sort(roomai.common.PokerCard.compare)

        num2point = [[], [], [], [], []]
        for p in pointrank2cards:
            num = len(pointrank2cards[p])
            num2point[num].append(p)
        for i in xrange(5):
            num2point[num].sort()

        sorted_point = []
        for p in pointrank2cards:
            sorted_point.append(p)
        sorted_point.sort()

        ##straight_samesuit
        for s in suitrank2cards:
            if len(suitrank2cards[s]) >= 5:
                numStraight = 1
                for i in xrange(len(suitrank2cards[s]) - 2, -1, -1):
                    if suitrank2cards[s][i].get_point_rank() == suitrank2cards[s][i + 1].get_point_rank() - 1:
                        numStraight += 1
                    else:
                        numStraight = 1

                    if numStraight == 5:
                        pattern = AllCardsPattern["Straight_SameSuit"]
                        pattern[6] = suitrank2cards[s][i:i + 5]
                        return pattern

        ##4_1
        if len(num2point[4]) > 0:
            p4 = num2point[4][0]
            p1 = -1
            for i in xrange(len(sorted_point) - 1, -1, -1):
                if sorted_point[i] != p4:
                    p1 = sorted_point[i]
                    break
            pattern = AllCardsPattern["4_1"]
            pattern[6] = pointrank2cards[p4][0:4]
            pattern[6].append(pointrank2cards[p1][0])
            return pattern

        ##3_2
        if len(num2point[3]) >= 1:
            pattern = AllCardsPattern["3_2"]

            if len(num2point[3]) == 2:
                p3 = num2point[3][1]
                pattern[6] = pointrank2cards[p3][0:3]
                p2 = num2point[3][0]
                pattern[6].append(pointrank2cards[p2][0])
                pattern[6].append(pointrank2cards[p2][1])
                return pattern

            if len(num2point[2]) >= 1:
                p3 = num2point[3][0]
                pattern[6] = pointrank2cards[p3][0:3]
                p2 = num2point[2][len(num2point[2]) - 1]
                pattern[6].append(pointrank2cards[p2][0])
                pattern[6].append(pointrank2cards[p2][1])
                return pattern

        ##SameSuit
        for s in suitrank2cards:
            if len(suitrank2cards[s]) >= 5:
                pattern = AllCardsPattern["SameSuit"]
                len1 = len(suitrank2cards[s])
                pattern[6] = suitrank2cards[s][len1 - 5:len1]
                return pattern

        ##Straight_DiffSuit
        numStraight = 1
        for idx in xrange(len(sorted_point) - 2, -1, -1):
            if sorted_point[idx] + 1 == sorted_point[idx]:
                numStraight += 1
            else:
                numStraight = 1

            if numStraight == 5:
                pattern = AllCardsPattern["Straight_DiffSuit"]
                for p in xrange(idx, idx + 5):
                    point = sorted_point[p]
                    pattern[6].append(pointrank2cards[point][0])
                return pattern

        ##3_1_1
        if len(num2point[3]) == 1:
            pattern = AllCardsPattern["3_1_1"]

            p3 = num2point[3][0]
            pattern[6] = pointrank2cards[p3][0:3]

            num = 0
            for i in xrange(len(sorted_point) - 1, -1, -1):
                p = sorted_point[i]
                if p != p3:
                    pattern[6].append(pointrank2cards[p][0])
                    num += 1
                if num == 2:    break
            return pattern

        ##2_2_1
        if len(num2point[2]) >= 2:
            pattern = AllCardsPattern["2_2_1"]
            p21 = num2point[2][len(num2point[2]) - 1]
            for c in pointrank2cards[p21]:
                pattern[6].append(c)
            p22 = num2point[2][len(num2point[2]) - 2]
            for c in pointrank2cards[p22]:
                pattern[6].append(c)

            flag = False
            for i in xrange(len(sorted_point) - 1, -1, -1):
                p = sorted_point[i]
                if p != p21 and p != p22:
                    c = pointrank2cards[p][0]
                    pattern[6].append(c)
                    flag = True
                if flag == True:    break;
            return pattern

        ##2_1_1_1
        if len(num2point[2]) == 1:
            pattern = AllCardsPattern["2_1_1_1"]
            p2 = num2point[2][0]
            pattern[6] = pointrank2cards[p2][0:2]
            num = 0
            for p in xrange(len(sorted_point) - 1, -1, -1):
                p1 = sorted_point[p]
                if p1 != p2:
                    pattern[6].append(pointrank2cards[p1][0])
                if num == 3:    break
            return pattern

        ##1_1_1_1_1
        pattern = AllCardsPattern["1_1_1_1_1"]
        count = 0
        for i in xrange(len(sorted_point) - 1, -1, -1):
            p = sorted_point[i]
            for c in pointrank2cards[p]:
                pattern[6].append(c)
                count += 1
                if count == 5: break
            if count == 5: break
        return pattern

    @classmethod
    def compare_handcards(cls, hand_card0, hand_card1, keep_cards):
        """

        Args:
            hand_card0:
            hand_card1:
            keep_cards:

        Returns:

        """
        pattern0 = TexasHoldemEnv.cards2pattern(hand_card0, keep_cards)
        pattern1 = TexasHoldemEnv.cards2pattern(hand_card1, keep_cards)

        diff = cls.compare_patterns(pattern0, pattern1)
        return diff

    @classmethod
    def compare_patterns(cls, p1, p2):
        """

        Args:
            p1:
            p2:

        Returns:

        """
        if p1[5] != p2[5]:
            return p1[5] - p2[5]
        else:
            for i in xrange(5):
                if p1[6][i] != p2[6][i]:
                    return p1[6][i] - p2[6][i]
            return 0

    @classmethod
    def available_actions(cls, public_state, person_state):
        """

        Args:
            public_state:
            person_state:

        Returns:

        """
        pu = public_state
        turn = pu.turn
        key_actions = dict()

        ## for fold
        action = TexasHoldemAction(TexasHoldemAction.Fold + "_0")
        if cls.is_action_valid(action,public_state, person_state):
            key_actions[action.key] = action

        ## for check
        if pu.bets[turn] == pu.max_bet_sofar:
            action = TexasHoldemAction(TexasHoldemAction.Check + "_0")
            if cls.is_action_valid(action, public_state, person_state):
                key_actions[action.key] = action

        ## for call
        if pu.bets[turn] != pu.max_bet_sofar and pu.chips[turn] > pu.max_bet_sofar - pu.bets[turn]:
            action = TexasHoldemAction(TexasHoldemAction.Call + "_%d" % (pu.max_bet_sofar - pu.bets[turn]))
            if cls.is_action_valid(action, public_state, person_state):
                key_actions[action.key] = action

        ## for raise
        if pu.bets[turn] != pu.max_bet_sofar and pu.chips[turn] > pu.max_bet_sofar - pu.bets[turn] + pu.raise_account:
            num = (pu.chips[turn] - (pu.max_bet_sofar - pu.bets[turn])) / pu.raise_account
            for i in xrange(1, num + 1):
                price = pu.max_bet_sofar - pu.bets[turn] + pu.raise_account * i
                if price == pu.chips[pu.turn]:  continue
                action = TexasHoldemAction(TexasHoldemAction.Raise + "_%d" % (price))
                if cls.is_action_valid(action, public_state, person_state):
                    key_actions[action.key] = action

        ## for all in
        action = TexasHoldemAction(TexasHoldemAction.AllIn + "_%d" % (pu.chips[turn]))
        if cls.is_action_valid(action, public_state, person_state):
            key_actions[action.key] = action

        return key_actions

    @classmethod
    def is_action_valid(cls, action, public_state, person_state):
        """

        Args:
            action:
            public_state:
            person_state:

        Returns:

        """
        pu = public_state

        if (not isinstance(public_state, TexasHoldemPublicState)) or (not isinstance(action, TexasHoldemAction)):
            return False

        if pu.is_allin[pu.turn] == True or pu.is_fold[pu.turn] == True:
            return False
        if pu.chips[pu.turn] == 0:
            return False

        if action.option == TexasHoldemAction.Fold:
            return True

        elif action.option == TexasHoldemAction.Check:
            if pu.bets[pu.turn] == pu.max_bet_sofar:
                return True
            else:
                return False

        elif action.option == TexasHoldemAction.Call:
            if action.price == pu.max_bet_sofar - pu.bets[pu.turn]:
                return True
            else:
                return False

        elif action.option == TexasHoldemAction.Raise:
            raise_account = action.price - (pu.max_bet_sofar - pu.bets[pu.turn])
            if raise_account == 0:    return False
            if raise_account % pu.raise_account == 0:
                return True
            else:
                return False
        elif action.option == TexasHoldemAction.AllIn:
            if action.price == pu.chips[pu.turn]:
                return True
            else:
                return False
        else:
            raise Exception("Invalid action.option" + action.option)




#!/bin/python
#coding:utf-8

import random
import copy
import itertools

from roomai.doudizhu.DouDiZhuPokerInfo   import *
from roomai.doudizhu.DouDiZhuPokerAction import *

class DouDiZhuPokerEnv(roomai.common.AbstractEnv):
    """
    """

    def __init__(self):
        """

        """
        self.public_state  = DouDiZhuPublicState()
        self.private_state = DouDiZhuPrivateState()
        self.person_states = [DouDiZhuPersonState() for i in range(3)]



    def update_license(self, turn, action):
        """

        Args:
            turn:
            action:
        """
        if action.pattern[0] != "i_cheat":
            self.public_state.license_playerid = turn
            self.public_state.license_action   = action 
            

    def update_cards(self, turn, action):
        """

        Args:
            turn:
            action:
        """
        self.person_states[turn].hand_cards.remove_action(action)


    def update_phase_bid2play(self):
        """

        """
        self.public_state.phase                 = 1
        
        self.public_state.landlord_id           = self.public_state.landlord_candidate_id
        self.public_state.license_playerid      = self.public_state.turn
        self.public_state.continuous_cheat_num  = 0
        self.public_state.is_response           = False

        landlord_id = self.public_state.landlord_id
        self.public_state.keep_cards = DouDiZhuHandCards(self.private_state.keep_cards.key)
        self.person_states[landlord_id].hand_cards.add_cards(self.private_state.keep_cards)


    #@Overide
    def init(self, params = dict()):
        """

        Args:
            params:

        Returns:

        """

        if "allcards" in params:
            self.allcards = [c for c in params["allcards"]]
        else:
            self.cards = []
            for i in range(13):
                for j in range(4):
                    self.cards.append(DouDiZhuActionElement.rank_to_str[i])
            self.cards.append(DouDiZhuActionElement.rank_to_str[13])
            self.cards.append(DouDiZhuActionElement.rank_to_str[14])
            random.shuffle(self.cards)

        if "record_history" in params:
            self.record_history = params["record_history"]
        else:
            self.record_history = False

        if "start_turn" in params:
            self.start_turn = params["start_turn"]
        else:
            self.start_turn = int(random.random() * 2)

        for i in range(3):
            tmp = self.cards[i*17:(i+1)*17]
            tmp.sort()
            self.person_states[i].hand_cards = DouDiZhuHandCards("".join(tmp))

        keep_cards = DouDiZhuHandCards([self.cards[-1], self.cards[-2], self.cards[-3]])
        self.private_state.keep_cards =  keep_cards;
        
        self.public_state.firstPlayer         = self.start_turn
        self.public_state.turn                = self.public_state.firstPlayer
        self.public_state.phase               = 0
        self.public_state.epoch               = 0
        
        self.public_state.landlord_id         = -1
        self.public_state.license_playerid    = self.public_state.turn
        self.public_state.license_action      = None
        self.public_state.is_terminal         = False
        self.public_state.scores              = [0,0,0]

        turn = self.public_state.turn
        self.person_states[turn].available_actions = DouDiZhuPokerEnv.available_actions(self.public_state, self.person_states[turn])

        infos = self.__gen_infos__()
        self.__gen_history__()

        return infos, self.public_state, self.person_states, self.private_state


    ## we need ensure the action is valid
    #@Overide
    def forward(self, action):
        """

        Args:
            action:

        Returns:

        """

        if self.is_action_valid(action, self.public_state, self.person_states[self.public_state.turn]) is False:
            raise  ValueError("%s action is invalid"%(action.key))

        turn = self.public_state.turn
        if self.public_state.phase == 0:
            if action.pattern[0] == "i_bid":
                self.public_state.landlord_candidate_id = turn

            if self.public_state.epoch == 3 and self.public_state.landlord_candidate_id == -1:
                self.public_state.is_terminal = True
                self.public_state.scores      = [0.0, 0.0, 0.0]

                self.__gen_history__()
                infos = self.__gen_infos__()
                return infos, self.public_state, self.person_states, self.private_state

            if (self.public_state.epoch == 2 and self.public_state.landlord_candidate_id != -1)\
                or self.public_state.epoch == 3:
                self.update_phase_bid2play()


                self.public_state.previous_id = turn
                self.public_state.previous_action = action
                self.public_state.epoch += 1
                self.person_states[self.public_state.turn].available_actions = DouDiZhuPokerEnv.available_actions(
                    self.public_state, self.person_states[self.public_state.turn])

                self.__gen_history__()
                infos = self.__gen_infos__()

                return infos, self.public_state, self.person_states, self.private_state


        else: #phase == play

            if action.pattern[0] != "i_cheat":
                
                self.update_cards(turn,action)
                self.update_license(turn,action)
                self.public_state.continuous_cheat_num = 0
    
                num = self.person_states[turn].hand_cards.num_cards
                if num == 0:
                    self.public_state.previous_id = turn
                    self.public_state.previous_action = action
                    self.public_state.epoch += 1
                    if turn == self.public_state.landlord_id:
                        self.public_state.is_terminal                           = True
                        self.public_state.scores                                = [-1,-1,-1]
                        self.public_state.scores[self.public_state.landlord_id] = 2
                    else:
                        self.public_state.is_terminal                           = True
                        self.public_state.scores                                = [1,1,1]
                        self.public_state.scores[self.public_state.landlord_id] = -2
                    self.__gen_history__()
                    infos = self.__gen_infos__()
                    return infos, self.public_state, self.person_states, self.private_state
            else:
                self.public_state.continuous_cheat_num += 1


        self.public_state.turn   = (turn+1)%3


        if self.public_state.continuous_cheat_num == 2:
            self.public_state.is_response          = False
            self.public_state.continuous_cheat_num = 0
        else:
            self.public_state.is_response = True


        self.public_state.previous_id         = turn
        self.public_state.previous_action     = action
        self.public_state.epoch              += 1
        self.person_states[self.public_state.turn].available_actions = DouDiZhuPokerEnv.available_actions(self.public_state, self.person_states[self.public_state.turn])
         
        self.__gen_history__()
        infos = self.__gen_infos__()

        return infos, self.public_state, self.person_states, self.private_state


    #@override
    @classmethod
    def compete(cls, env, players):
        """

        Args:
            env:
            players:

        Returns:

        """
        infos ,public_state, person_states, private_state= env.init()

        for i in range(len(players)):
            players[i].receive_info(infos[i])

        while public_state.is_terminal == False:
            turn = public_state.turn
            action = players[turn].takeAction()
            infos, public_state, person_states, private_state = env.forward(action)
            for i in range(len(players)):
                players[i].receive_info(infos[i])

        return public_state.scores



    @classmethod
    def available_actions(cls, public_state, person_state):
        """

        Args:
            public_state:
            person_state:

        Returns:

        """

        patterns = []
        if public_state.phase == 0:
            patterns.append(AllPatterns["i_cheat"])
            patterns.append(AllPatterns["i_bid"])
        else:
            if public_state.is_response == False:
                for p in AllPatterns:
                    if p != "i_cheat" and p != "i_invalid":
                        patterns.append(AllPatterns[p])
            else:
                patterns.append(public_state.license_action.pattern)
                if public_state.license_action.pattern[6] == 1:
                    patterns.append(AllPatterns["p_4_1_0_0_0"])  # rank = 10
                    patterns.append(AllPatterns["x_rocket"])  # rank = 100
                if public_state.license_action.pattern[6] == 10:
                    patterns.append(AllPatterns["x_rocket"])  # rank = 100
                patterns.append(AllPatterns["i_cheat"])

        is_response = public_state.is_response
        license_act = public_state.license_action
        actions = dict()

        for pattern in patterns:
            numMaster = pattern[1]
            numMasterPoint = pattern[2]
            isStraight = pattern[3]
            numSlave = pattern[4]
            MasterCount = -1
            SlaveCount = -1

            if numMaster > 0:
                MasterCount = int(numMaster / numMasterPoint)

            if "i_invalid" == pattern[0]:
                continue

            if "i_cheat" == pattern[0]:
                action_key = DouDiZhuPokerAction.master_slave_cards_to_key([DouDiZhuActionElement.cheat], [])
                action     = DouDiZhuPokerAction.lookup(action_key)
                if cls.is_action_valid(action,public_state, person_state) == True:
                    actions[action_key] = action
                continue

            if "i_bid" == pattern[0]:
                action_key = DouDiZhuPokerAction.master_slave_cards_to_key([DouDiZhuActionElement.bid], [])
                action     = DouDiZhuPokerAction.lookup(action_key)
                if cls.is_action_valid(action, public_state,person_state) == True:
                    actions[action_key] = action
                continue

            if pattern[0] == "x_rocket":
                if person_state.hand_cards.cards[DouDiZhuActionElement.r] == 1 and \
                                person_state.hand_cards.cards[DouDiZhuActionElement.R] == 1:
                    action_key  = DouDiZhuPokerAction.master_slave_cards_to_key([DouDiZhuActionElement.r, DouDiZhuActionElement.R], [])
                    action      = DouDiZhuPokerAction.lookup(action_key)
                    if cls.is_action_valid(action,public_state,person_state) == True:
                        actions[action_key] = action
                continue

            if pattern[1] + pattern[4] > person_state.hand_cards.num_cards:
                continue
            sum1 = 0

            for count in range(MasterCount, 5, 1):
                sum1 += person_state.hand_cards.count2num[count]
            if sum1 < numMasterPoint:
                continue

            ### action with cards
            mCardss = []
            mCardss = DouDiZhuPokerEnv.extractMasterCards(person_state.hand_cards,  pattern)

            for mCards in mCardss:
                if numSlave == 0:
                    action_key   = DouDiZhuPokerAction.master_slave_cards_to_key(mCards, [])
                    action       = DouDiZhuPokerAction.lookup(action_key)
                    if cls.is_action_valid(action, public_state,person_state) == True:
                        actions[action_key] = action
                    continue

                sCardss = DouDiZhuPokerEnv.extractSlaveCards(person_state.hand_cards, mCards, pattern)
                for sCards in sCardss:
                    action_key  = DouDiZhuPokerAction.master_slave_cards_to_key(mCards, sCards)
                    action      = DouDiZhuPokerAction.lookup(action_key)
                    if cls.is_action_valid(action, public_state,person_state) == True:
                        actions[action_key] = action
        return actions



    @classmethod
    def is_action_valid(cls,action, public_state, person_state):
        '''
        print "is_action_valid_______________________________________________________________"
        print public_state.turn
        print person_state.hand_cards.num_cards
        print person_state.hand_cards.key
        print action.key
        '''

        if action.pattern[0] == "i_invalid":
            return False

        if cls.is_action_from_handcards(person_state.hand_cards, action) == False:
            return False

        turn        = public_state.turn
        license_id  = public_state.license_playerid
        license_act = public_state.license_action
        phase       = public_state.phase

        if phase == 0:
            if action.pattern[0] not in ["i_cheat", "i_bid"]:
                return False
            return True

        if phase == 1:
            if action.pattern[0] == "i_bid":    return False

            if public_state.is_response == False:
                if action.pattern[0] == "i_cheat": return False
                return True

            else:  # response
                if action.pattern[0] == "i_cheat":  return True
                ## not_cheat
                if action.pattern[6] > license_act.pattern[6]:
                    return True
                elif action.pattern[6] < license_act.pattern[6]:
                    return False
                elif action.maxMasterPoint - license_act.maxMasterPoint > 0:
                    return True
                else:
                    return False

    @classmethod
    def is_action_from_handcards(cls, hand_cards, action):
        """

        Args:
            hand_cards:
            action:

        Returns:

        """
        flag = True
        if action.pattern[0] == "i_cheat":  return True
        if action.pattern[0] == "i_bid":    return True
        if action.pattern[0] == "i_invalid":    return False

        for a in action.masterPoints2Count:
            flag = flag and (action.masterPoints2Count[a] <= hand_cards.cards[a])
        for a in action.slavePoints2Count:
            flag = flag and (action.slavePoints2Count[a] <= hand_cards.cards[a])
        return flag


    @classmethod
    def extractMasterCards(cls, hand_cards, pattern):
        """

        Args:
            hand_cards:
            pattern:

        Returns:

        """
        is_straight = pattern[3]
        cardss = []
        ss = []

        numPoint = pattern[2]
        if numPoint <= 0:
            return cardss
        count = pattern[1]/numPoint

        if is_straight == 1:
            c = 0
            for i in range(11, -1, -1):
                if hand_cards.cards[i] >= count:
                    c += 1
                else:
                    c = 0

                if c >= numPoint:
                    ss.append(range(i, i + numPoint))
        else:
            candidates = []
            for c in range(len(hand_cards.cards)):
                if hand_cards.cards[c] >= count:
                    candidates.append(c)
            if len(candidates) < numPoint:
                return []
            ss = list(itertools.combinations(candidates, numPoint))

        for set1 in ss:
            s = []
            for c in set1:
                for i in range(count):
                    s.append(c)
            s.sort()
            cardss.append(s)

        return cardss

    @classmethod
    def extractSlaveCards(cls, hand_cards, used_cards, pattern):
        """

        Args:
            hand_cards:
            used_cards:
            pattern:

        Returns:

        """
        used = [0 for i in range(15)]
        for p in used_cards:
            used[p] += 1

        numMaster = pattern[1]
        numMasterPoint = pattern[2]
        numSlave = pattern[4]

        candidates = []
        res1 = []
        res = []

        if numMaster / numMasterPoint == 3:
            if numSlave / numMasterPoint == 1:  # single
                for c in range(len(hand_cards.cards)):
                    for i in range(hand_cards.cards[c] - used[c]):
                        candidates.append(c)
                if len(candidates) >= numSlave:
                    res1 = list(set(list(itertools.combinations(candidates, numSlave))))
                for sCard in res1:  res.append([x for x in sCard])

            elif numSlave / numMasterPoint == 2:  # pair
                for c in range(len(hand_cards.cards)):
                    for i in range ((hand_cards.cards[c] - used[c])/2) :
                        candidates.append(c)
                if len(candidates) >= numSlave / 2:
                    res1 = list(set(list(itertools.combinations(candidates, int(numSlave / 2)))))
                for sCard in res1:
                    tmp = [x for x in sCard]
                    tmp.extend([x for x in sCard])
                    res.append(tmp)

        elif numMaster / numMasterPoint == 4:

            if numSlave / numMasterPoint == 2:  # single
                for c in range(len(hand_cards.cards)):
                    for i in range(hand_cards.cards[c] - used[c]):
                        candidates.append(c)
                if len(candidates) >= numSlave:
                    res1 = list(set(list(itertools.combinations(candidates, numSlave))))
                for sCard in res1:  res.append([x for x in sCard])


            elif numSlave / numMasterPoint == 4:  # pair
                for c in range(len(hand_cards.cards)):
                    for i in range((hand_cards.cards[c] - used[c])/2):
                        candidates.append(c)
                if len(candidates) >= numSlave / 2:
                    res1 = list(set(list(itertools.combinations(candidates, int(numSlave / 2)))))
                for sCard in res1:
                    tmp = [x for x in sCard]
                    tmp.extend([x for x in sCard])
                    res.append(tmp)

        return res


    @classmethod
    def action_priority(cls,action1, action2):
        """

        Args:
            action1:
            action2:

        Returns:

        """
        count1 = action1.pattern[1] / action1.pattern[2]
        count2 = action2.pattern[1] / action2.pattern[2]
        if count1 != count2:
            return count1 - count2

        numMaster1 = action1.pattern[1]
        numMaster2 = action2.pattern[2]
        if numMaster1 != numMaster2:
            return numMaster1  - numMaster2

        if action1.maxMasterPoint != action2.maxMasterPoint:
            return action1.maxMasterPoint - action2.maxMasterPoint

        raise ValueError("can't compare priorities of %s and %s "%(action1.key,action2))


    @classmethod
    def available_actions_generate_all(cls):
        """

        Returns:

        """
        public_state = DouDiZhuPublicState()
        person_state = DouDiZhuPersonState()
        public_state.is_response    = False
        person_state.hand_cards     = DouDiZhuHandCards("")
        for i in range(13):
            for j in range(4):
                person_state.hand_cards.add_cards(DouDiZhuActionElement.rank_to_str[i])
        person_state.hand_cards.add_cards(DouDiZhuActionElement.rank_to_str[DouDiZhuActionElement.r])
        person_state.hand_cards.add_cards(DouDiZhuActionElement.rank_to_str[DouDiZhuActionElement.R])
        actions = dict()


        patterns = []
        if public_state.phase == 0:
            patterns.append(AllPatterns["i_cheat"])
            patterns.append(AllPatterns["i_bid"])
        else:
            if public_state.is_response == False:
                for p in AllPatterns:
                    if p != "i_cheat" and p != "i_invalid":
                        patterns.append(AllPatterns[p])
            else:
                patterns.append(public_state.license_action.pattern)
                if public_state.license_action.pattern[6] == 1:
                    patterns.append(AllPatterns["p_4_1_0_0_0"])  # rank = 10
                    patterns.append(AllPatterns["x_rocket"])  # rank = 100
                if public_state.license_action.pattern[6] == 10:
                    patterns.append(AllPatterns["x_rocket"])  # rank = 100
                patterns.append(AllPatterns["i_cheat"])


        actions = dict()

        for pattern in patterns:
            numMaster = pattern[1]
            numMasterPoint = pattern[2]
            isStraight = pattern[3]
            numSlave = pattern[4]
            MasterCount = -1
            SlaveCount = -1

            if numMaster > 0:
                MasterCount = int(numMaster / numMasterPoint)

            if "i_invalid" == pattern[0]:
                continue

            if "i_cheat" == pattern[0]:
                action_key = DouDiZhuPokerAction.master_slave_cards_to_key([DouDiZhuActionElement.cheat], [])
                action     = DouDiZhuPokerAction([DouDiZhuActionElement.cheat], [])
                if action_key in actions:
                    if cls.action_priority(action, actions[action_key]) > 0:
                        actions[action_key] = action
                else:
                    actions[action_key] = action

                continue

            if "i_bid" == pattern[0]:
                action_key = DouDiZhuPokerAction.master_slave_cards_to_key([DouDiZhuActionElement.bid], [])
                action = DouDiZhuPokerAction([DouDiZhuActionElement.bid], [])
                if action_key in actions:
                    if cls.action_priority(action, actions[action_key]) > 0:
                        actions[action_key] = action
                else:
                    actions[action_key] = action
                continue

            if pattern[0] == "x_rocket":
                if person_state.hand_cards.cards[DouDiZhuActionElement.r] == 1 and \
                                person_state.hand_cards.cards[DouDiZhuActionElement.R] == 1:
                    action_key  = DouDiZhuPokerAction.master_slave_cards_to_key([DouDiZhuActionElement.r, DouDiZhuActionElement.R], [])
                    action = DouDiZhuPokerAction([DouDiZhuActionElement.r, DouDiZhuActionElement.R], [])
                    if action_key in actions:
                        if cls.action_priority(action, actions[action_key]) > 0:
                            actions[action_key] = action
                    else:
                        actions[action_key] = action
                continue

            if pattern[1] + pattern[4] > person_state.hand_cards.num_cards:
                continue
            sum1 = 0

            for count in range(MasterCount, 5, 1):
                sum1 += person_state.hand_cards.count2num[count]
            if sum1 < numMasterPoint:
                continue

            ### action with cards
            mCardss = []
            mCardss = DouDiZhuPokerEnv.extractMasterCards(person_state.hand_cards,  pattern)

            for mCards in mCardss:
                if numSlave == 0:
                    action_key   = DouDiZhuPokerAction.master_slave_cards_to_key(mCards, [])
                    action = DouDiZhuPokerAction(mCards, [])
                    if action_key in actions:
                        if cls.action_priority(action, actions[action_key]) > 0:
                            actions[action_key] = action
                    else:
                        actions[action_key] = action
                    continue

                sCardss = DouDiZhuPokerEnv.extractSlaveCards(person_state.hand_cards, mCards, pattern)
                for sCards in sCardss:
                    action_key  = DouDiZhuPokerAction.master_slave_cards_to_key(mCards, sCards)
                    action = DouDiZhuPokerAction(mCards, sCards)
                    if action_key in actions:
                        if cls.action_priority(action, actions[action_key]) > 0:
                            actions[action_key] = action
                    else:
                        actions[action_key] = action
        return actions



#!/bin/python
import roomai.common

class SevenKingPublicState(roomai.common.AbstractPublicState):
    """
    """
    def __init__(self):
        """

        """
        super(SevenKingPublicState,self).__init__()
        self.stage            = None
        self.num_players      = None
        self.showed_cards     = None
        self.num_showed_cards = None
        self.num_keep_cards   = None
        self.num_hand_cards   = None
        self.is_fold          = None
        self.num_fold         = None
        self.license_action   = None

    def __deepcopy__(self, newinstance = None, memodict={}):
        """

        Args:
            newinstance:
            memodict:

        Returns:

        """
        if  newinstance is None:
            newinstance = SevenKingPublicState()
        newinstance            = super(SevenKingPublicState,self).__deepcopy__(newinstance = newinstance)

        if self.showed_cards is None:
            newinstance.showed_cards = None
        else:
            newinstance.showed_cards = [card.__deepcopy__() for card in self.showed_cards]
        return newinstance

class SevenKingPrivateState(roomai.common.AbstractPrivateState):
    """
    """
    def __init__(self):
        """

        """
        super(SevenKingPrivateState,self).__init__()
        self.keep_cards   = []

    def __deepcopy__(self, newinstance = None, memodict={}):
        """

        Args:
            newinstance:
            memodict:

        Returns:

        """
        if newinstance is None:
            newinstance = SevenKingPrivateState()
        newinstance            = super(SevenKingPrivateState,self).__deepcopy__(newinstance = newinstance)
        newinstance.keep_cards =  [card.__deepcopy__() for card in self.keep_cards   ]
        return newinstance


class SevenKingPersonState(roomai.common.AbstractPersonState):
    """
    """
    def __init__(self):
        """

        """
        super(SevenKingPersonState,self).__init__()
        self.hand_cards   = []

    def __deepcopy__(self, memodict={}, newinstance = None):
        """

        Args:
            memodict:
            newinstance:

        Returns:

        """
        if newinstance is None:
            newinstance        = SevenKingPersonState()
        newinstance            = super(SevenKingPersonState, self).__deepcopy__(newinstance= newinstance)
        newinstance.hand_cards = [card.__deepcopy__() for card in self.hand_cards]
        return newinstance






#!/bin/python
#coding:utf-8
import roomai.common


class FiveCardStudAction(roomai.common.AbstractAction):
    """
    """

    # 弃牌
    Fold        = "Fold"
    # 过牌
    Check       = "Check"
    # 更注
    Call        = "Call"
    # 加注
    Raise       = "Raise"
    # 下注
    Bet         = "Bet"
    # all in
    Showhand    = "Showhand"

    def __init__(self,key):
            """

            Args:
                key:
            """
            super(FiveCardStudAction,self).__init__(key)
            opt_price = key.strip().split("_")
            if  opt_price[0] != self.Fold    and opt_price[0] != self.Call  and \
                opt_price[0] != self.Check   and opt_price[0] != self.Raise and \
                opt_price[0] != self.Bet     and opt_price[0] != self.Showhand:
                raise  ValueError("%s is an invalid key. The Option must be in [Fold,Check,Call,Raise,Bet,Showhand]"%key)

            if opt_price[0] not in  [self.Fold,self.Check, self.Call] and int(opt_price[1]) <= 0:
                raise  ValueError("%s is an invalid key.]"%key)

            if opt_price[0] == self.Fold and int(opt_price[1]) > 0:
                raise  ValueError("%s is an invalid key"%(key))

            if int(opt_price[1]) < 0:
                raise  ValueError("%s is an invalid key.]"%key)

            self._roomai_option = opt_price[0]
            self._roomai_price  = int(opt_price[1])



    @property
    def option(self):
        """

        Returns:

        """
        return self._roomai_option
    @property
    def price(self):
        """

        Returns:

        """
        return self._roomai_price

    @classmethod
    def lookup(cls,key):
        """

        Args:
            key:

        Returns:

        """
        return AllFiveCardStudActions[key]

    def __deepcopy__(self, memodict={}, newinstance = None):
        """

        Args:
            memodict:
            newinstance:

        Returns:

        """
        if newinstance is None:
            newinstance        = AllFiveCardStudActions[self.key]
        return newinstance

AllFiveCardStudActions = dict()
options = ["Fold", "Check","Call","Raise","Bet","Showhand"]
for option in options:
    if option in ["Check","Call"]:
        for p in range(100000):
            AllFiveCardStudActions["%s_%d"%(option,p)] = FiveCardStudAction("%s_%d"%(option,p))
    elif option in ["Raise","Bet","Showhand"]:
        for p in range(1,100000):
            AllFiveCardStudActions["%s_%d"%(option,p)] = FiveCardStudAction("%s_%d"%(option,p))
    else:
        AllFiveCardStudActions["Fold_0"] = FiveCardStudAction("Fold_0")
is_init_action = False



#!/bin/python
import roomai.common
import roomai.sevenking
from roomai.sevenking import AllSevenKingPatterns


class SevenKingAction(roomai.common.AbstractAction):
    """
    """

    def __init__(self, key):
        """

        Args:
            key:
        """
        if not isinstance(key,str):
            raise TypeError("The key for SevenKingAction is an str, not %s"%(type(str)))

        super(SevenKingAction,self).__init__(key)
        self.__key         = key.strip()
        self.__cards       = []
        if len(key) > 0:
            for c in self.key.split(","):
                self.__cards.append(roomai.sevenking.SevenKingPokerCard(c))
            self.__cards.sort(cmp = roomai.sevenking.SevenKingPokerCard.compare)
        self.__pattern = self.action2pattern(self)

    @classmethod
    def action2pattern(cls, action):
        """

        Args:
            action:

        Returns:

        """
        ###numCards
        num_cards  = len(action.cards)
        return AllSevenKingPatterns["p_%d"%(num_cards)]

    @property
    def key(self):
        """

        Returns:

        """
        return self.__key

    @property
    def cards(self):
        """

        Returns:

        """
        return self.__cards

    @property
    def pattern(self):
        """

        Returns:

        """
        return self.__pattern

    @classmethod
    def lookup(cls, key):
        """

        Args:
            key:

        Returns:

        """
        if key in AllSevenKingActions:
            return AllSevenKingActions[key]
        else:
            AllSevenKingActions[key] = SevenKingAction(key)
            return AllSevenKingActions[key]

    def __deepcopy__(self, memodict={}, newinstance = None):
        """

        Args:
            memodict:
            newinstance:

        Returns:

        """
        if newinstance is None:
            newinstance = SevenKingAction(self.key)
        newinstance         = super(SevenKingAction,self).__deepcopy__(newinstance = newinstance)
        newinstance.__key     = self.key
        newinstance.__cards   = [card.__deepcopy__() for card in self.cards]
        newinstance.__pattern = self.pattern
        return newinstance

AllSevenKingActions = dict()



#!/bin/python
import roomai.common

                                                    #0     1           2       3           4                                    5
                                                    #name, isStraight, isSameSuit, isNumRelated, [SizeOfPair1, SizeOfPair2,..](desc), rank
FiveCardStudAllCardsPattern = dict()
FiveCardStudAllCardsPattern["Straight_SameSuit"] = ["Straight_SameSuit", True, True, False, [], 100]
FiveCardStudAllCardsPattern["4_1"]               = ["4_1", False, False, True, [4, 1], 98]
FiveCardStudAllCardsPattern["3_2"]               = ["3_2", False, False, True, [3, 2], 97]
FiveCardStudAllCardsPattern["SameSuit"]          = ["SameSuit", False, True, False, [], 96]
FiveCardStudAllCardsPattern["Straight_DiffSuit"] = ["Straight_DiffSuit", True, False, False, [], 95]
FiveCardStudAllCardsPattern["3_1_1"]             = ["3_1_1", False, False, True, [3, 1, 1], 94]
FiveCardStudAllCardsPattern["2_2_1"]             = ["2_2_1", False, False, True, [2, 2, 1], 93]
FiveCardStudAllCardsPattern["2_1_1_1"]           = ["2_1_1_1", False, False, True, [2, 1, 1, 1], 92]
FiveCardStudAllCardsPattern["1_1_1_1_1"]         = ["1_1_1_1_1", False, False, True, [1, 1, 1, 1, 1], 91]


class FiveCardStudPokerCard(roomai.common.PokerCard):
    """
    """

    def get_suit_rank(self):
        """

        Returns:

        """
        suit_str_to_rank = {'Spade': 3, 'Heart': 2, 'Club': 1, 'Diamond':0}
        return suit_str_to_rank[self.suit_str]

    @classmethod
    def compare(cls, pokercard1, pokercard2):
        """

        Args:
            pokercard1:
            pokercard2:

        Returns:

        """
        pr1 = pokercard1.get_point_rank()
        pr2 = pokercard2.get_point_rank()
        if pr1 != pr2:
            return pr1-pr2
        else:
            return pokercard1.get_suit_rank() - pokercard2.get_suit_rank()


    def __deepcopy__(self, memodict={}, newinstance = None):
        """

        Args:
            memodict:
            newinstance:

        Returns:

        """
        if newinstance is None:
            newinstance = FiveCardStudAllPokerCards[self.key]
        return newinstance

    @classmethod
    def lookup(cls, key):
        """

        Args:
            key:

        Returns:

        """
        return FiveCardStudAllPokerCards[key]


FiveCardStudAllPokerCards = dict()
for point_str in roomai.common.common.point_str_to_rank:
    if point_str != 'r' and point_str != "R":
        for suit_str in roomai.common.common.suit_str_to_rank:
            if suit_str != "ForKing":
                FiveCardStudAllPokerCards["%s_%s"%(point_str,suit_str)] = FiveCardStudPokerCard("%s_%s"%(point_str,suit_str))
FiveCardStudAllPokerCards["r_ForKing"] = (FiveCardStudPokerCard("r_ForKing"))
FiveCardStudAllPokerCards["R_ForKing"] = (FiveCardStudPokerCard("R_ForKing"))



#!/bin/python
import roomai.sevenking
import random
random.seed(4)
class HumanInputPlayer(roomai.sevenking.SevenKingRandomPlayer):
    """
    """
    def receive_info(self, info):
        """

        Args:
            info:
        """
        available_actions = info
    def take_action(self):
        """

        Returns:

        """
        action = input("choosed_acton:")
        #action = ""
        return roomai.sevenking.SevenKingAction.lookup(action)
    def reset(self):
        """

        """
        pass

class HumanInputPlayer1(roomai.sevenking.SevenKingRandomPlayer):
    """
    """
    def receive_info(self, info):
        """

        Args:
            info:
        """
        available_actions = info
    def take_action(self):
        """

        Returns:

        """
        action = input("choosed_acton:")
        #action = ""
        return roomai.sevenking.SevenKingAction.lookup(action)
    def reset(self):
        """

        """
        pass

def show(info):
    """

    Args:
        info:
    """
    person_state          = info.person_state
    person_state.hand_cards.sort(cmp = roomai.sevenking.SevenKingPokerCard.compare)
    sorted_hand_cards_str = [c.key for c in person_state.hand_cards]
    print ("%s"%(person_state.id) + "'s hand_cards:\t" + ",".join(sorted_hand_cards_str))
    print ("%s"%(person_state.id) + "'s available_actions:\t" + " ".join(person_state.available_actions.keys()))

if __name__ == "__main__":
    players     = [HumanInputPlayer(), HumanInputPlayer1()]
    env         = roomai.sevenking.SevenKingEnv()
    allcards    = roomai.sevenking.AllSevenKingPokerCards.values()[0:17]
    allcards.sort(cmp = roomai.sevenking.SevenKingPokerCard.compare)

    tmp          = allcards[-6]
    allcards[-6] = allcards[-1]
    allcards[-1] = tmp

    tmp          = allcards[-7]
    allcards[-7] = allcards[-2]
    allcards[-2] = tmp


    num_players = len(players)
    infos, public_state, person_states, private_state = env.init({"num_players": num_players,"allcards":allcards})
    for i in range(env.num_players):
        players[i].receive_info(infos[i])
        show(infos[i])
    print (public_state.is_fold)
    print "\n"


    while public_state.is_terminal == False:
        turn = public_state.turn
        print "turn = %d, stage = %d"%(public_state.turn,public_state.stage)
        action = players[turn].take_action()
        print "%d player take an action (%s)"%(turn,action.key)
        infos, public_state, person_states, private_state = env.forward(action)
        for i in range(env.num_players):
            players[i].receive_info(infos[i])
            show(infos[i])
        print (public_state.is_fold)
        print "\n"

    print (public_state.scores)



#!/bin/python
import os
import roomai.common
import copy


#
#0, 1, 2, 3, ..., 7,  8, 9, 10, 11, 12, 13, 14
#^                ^   ^              ^       ^
#|                |   |              |       |
#3,               10, J, Q,  K,  A,  2,  r,  R
#

class DouDiZhuActionElement:
    """
    """
    str_to_rank  = {'3':0, '4':1, '5':2, '6':3, '7':4, '8':5, '9':6, 'T':7, 'J':8, 'Q':9, 'K':10, 'A':11, '2':12, 'r':13, 'R':14, 'x':15, 'b':16}
    # x means check, b means bid
    rank_to_str  = {0: '3', 1: '4', 2: '5', 3: '6', 4: '7', 5: '8', 6: '9', 7: 'T', 8: 'J', 9: 'Q', 10: 'K', 11: 'A', 12: '2', 13: 'r', 14: 'R', 15: 'x', 16: 'b'}
    three       = 0;
    four        = 1;
    five        = 2;
    six         = 3;
    seven       = 4;
    eight       = 5;
    night       = 6;
    ten         = 7;
    J           = 8;
    Q           = 9;
    K           = 10;
    A           = 11;
    two         = 12;
    r           = 13;
    R           = 14;
    cheat       = 15;
    bid         = 16;

    total_normal_cards = 15



class DouDiZhuPokerAction(roomai.common.AbstractAction):
    """
    """
    def __init__(self):
        """

        """
        pass

    def __init__(self, masterCards, slaveCards):
        """

        Args:
            masterCards:
            slaveCards:
        """
        self.__masterCards        = copy.deepcopy(masterCards)
        self.__slaveCards         = copy.deepcopy(slaveCards)

        self.__masterPoints2Count = None
        self.__slavePoints2Count  = None
        self.__isMasterStraight   = None
        self.__maxMasterPoint     = None
        self.__minMasterPoint     = None
        self.__pattern            = None

        self.action2pattern()
        self.__key = DouDiZhuPokerAction.master_slave_cards_to_key(masterCards, slaveCards)



    @property
    def key(self):  return self.__key
    @property
    def masterCards(self):  return self.__masterCards
    @property
    def slaveCards(self):   return self.__slaveCards
    @property
    def masterPoints2Count(self):   return self.__masterPoints2Count
    @property
    def slavePoints2Count(self):    return self.__slavePoints2Count
    @property
    def isMasterStraight(self):     return self.__isMasterStraight
    @property
    def maxMasterPoint(self):       return self.__maxMasterPoint
    @property
    def minMasterPoint(self):       return self.__minMasterPoint
    @property
    def pattern(self):              return self.__pattern

    @classmethod
    def lookup(cls, key):
        """

        Args:
            key:

        Returns:

        """
        return AllActions["".join(sorted(key))]

    @classmethod
    def master_slave_cards_to_key(cls, masterCards, slaveCards):
        """

        Args:
            masterCards:
            slaveCards:

        Returns:

        """
        key_int = (masterCards + slaveCards)
        key_str = []
        for key in key_int:
            key_str.append(DouDiZhuActionElement.rank_to_str[key])
        key_str.sort()
        return "".join(key_str)

    def action2pattern(self):
        """

        """

        self.__masterPoints2Count = dict()
        for c in self.__masterCards:
            if c in self.__masterPoints2Count:
                self.__masterPoints2Count[c] += 1
            else:
                self.__masterPoints2Count[c] = 1

        self.__slavePoints2Count = dict()
        for c in self.__slaveCards:
            if c in self.__slavePoints2Count:
                self.__slavePoints2Count[c] += 1
            else:
                self.__slavePoints2Count[c] = 1

        self.__isMasterStraight = 0
        num = 0
        for v in self.__masterPoints2Count:
            if (v + 1) in self.__masterPoints2Count and (v + 1) < DouDiZhuActionElement.two:
                num += 1
        if num == len(self.__masterPoints2Count) - 1 and len(self.__masterPoints2Count) != 1:
            self.__isMasterStraight = 1

        self.__maxMasterPoint = -1
        self.__minMasterPoint = 100
        for c in self.__masterPoints2Count:
            if self.__maxMasterPoint < c:
                self.__maxMasterPoint = c
            if self.__minMasterPoint > c:
                self.__minMasterPoint = c

        ########################
        ## action 2 pattern ####
        ########################


        # is cheat?
        if len(self.__masterCards) == 1 \
                and len(self.__slaveCards) == 0 \
                and self.__masterCards[0] == DouDiZhuActionElement.cheat:
            self.__pattern = AllPatterns["i_cheat"]

        # is roblord
        elif len(self.__masterCards) == 1 \
                and len(self.__slaveCards) == 0 \
                and self.__masterCards[0] == DouDiZhuActionElement.bid:
            self.__pattern = AllPatterns["i_bid"]

        # is twoKings
        elif len(self.__masterCards) == 2 \
                and len(self.__masterPoints2Count) == 2 \
                and len(self.__slaveCards) == 0 \
                and self.__masterCards[0] in [DouDiZhuActionElement.r, DouDiZhuActionElement.R] \
                and self.__masterCards[1] in [DouDiZhuActionElement.r, DouDiZhuActionElement.R]:
            self.__pattern = AllPatterns["x_rocket"]

        else:

            ## process masterCards
            masterPoints = self.__masterPoints2Count
            if len(masterPoints) > 0:
                count = masterPoints[self.__masterCards[0]]
                for c in masterPoints:
                    if masterPoints[c] != count:
                        self.__pattern = AllPatterns["i_invalid"]

            if self.__pattern == None:
                pattern = "p_%d_%d_%d_%d_%d" % (len(self.__masterCards), len(masterPoints), \
                                                self.__isMasterStraight, \
                                                len(self.__slaveCards), 0)

                if pattern in AllPatterns:
                    self.__pattern = AllPatterns[pattern]
                else:
                    self.__pattern = AllPatterns["i_invalid"]

    def __deepcopy__(self, memodict={}, newinstance = None):
        """

        Args:
            memodict:
            newinstance:

        Returns:

        """
        return self.lookup(self.key)



############## read data ################
AllPatterns = dict()
AllActions = dict()
import zipfile
def get_file(path):
    """

    Args:
        path:

    Returns:

    """
    if ".zip" in path:
        lines = path.split(".zip")
        zip1 = zipfile.ZipFile(lines[0] + ".zip")
        len1 = len(lines[1])
        path = lines[1][1:len1]
        return zip1.open(path)
    else:
        return open(path)
path = os.path.split(os.path.realpath(__file__))[0]
pattern_file = get_file(path + "/patterns.py")
for line in pattern_file:
    line = line.replace(" ", "").strip()
    line = line.split("#")[0]
    if len(line) == 0:  continue
    lines = line.split(",")
    for i in range(1, len(lines)):
        lines[i] = int(lines[i])
    AllPatterns[lines[0]] = lines
pattern_file.close()



action_file = get_file(path + "/actions.py")
for line in action_file:
    line = line.replace(" ", "").strip()
    lines = line.split("\t")
    m = [int(str1) for str1 in lines[1].split(",")]
    s = []
    if len(lines[2]) > 0:
        s = [int(str1) for str1 in lines[2].split(",")]
    action = DouDiZhuPokerAction(m, s)
    if action.key != lines[0] or action.pattern[0] != lines[3]:
        print lines
        raise ValueError("%s is wrong. The generated action has key(%s) and pattern(%s)"%(line, action.key,action.pattern[0]))


    AllActions[action.key] = action
action_file.close()






#!/bin/python
import logging;
import sys;

project_name = "roomai";

logger = logging.getLogger(project_name);
handler = logging.StreamHandler(sys.stderr);
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s");

logger.setLevel(logging.INFO);
handler.setLevel(logging.INFO);
handler.setFormatter(formatter);
logger.addHandler(handler);


def set_level(level):
    """

    Args:
        level:
    """
    logger.setLevel(level)
    handler.setLevel(level)

def get_logger():
    """

    Returns:

    """
    return logger


def init_logger(opts):
    """

    Args:
        opts:
    """
    global logger;
    global handler;
    global project_name;

    print opts;
    if "project_name" in opts:
        project_name = opts["project_name"];
        print "in Logger", project_name;

    logger.removeHandler(handler);
    logger = logging.getLogger(project_name);

    #set longer
    if "logfile" in opts:
        handler = logging.FileHandler(opts["logfile"]);
    else:
        handler = logging.StreamHandler(sys.stderr);
    
    #set formatter
    formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s");
    handler.setFormatter(formatter);

    ##set level
    logger.setLevel(logging.INFO);
    if "level" in opts:
        if "notset" == opts["level"].lowcase():
            logger.setLevel(logging.NOTSET)
        elif "debug" == opts["level"].lowcase():
            logger.setLevel(logging.DEBUG)
        elif "info"  == opts["level"].lowcase():
            logger.setLevel(logging.INFO)
        elif "warning" == opts["level"].lowcase():
            logger.setLevel(logging.WARNING)
        elif "error" == opts["level"].lowcase():
            logger.setLevel(logging.ERROR)
        elif "critical" == opts["level"].lowcase():
            logger.setLevel(logging.critical)

    logger.addHandler(handler);
   



#!/bin/python
from common import AbstractPlayer
from common import AbstractEnv
from common import AbstractPublicState
from common import AbstractPersonState
from common import AbstractPrivateState
from common import AbstractAction
from common import PokerCard
from common import Info



#!/bin/python
import roomai.common
import copy


class FiveCardStudPrivateState(roomai.common.AbstractPrivateState):
    """
    """
    all_hand_cards    = None

    def __deepcopy__(self, memodict={}):
        """

        Args:
            memodict:

        Returns:

        """
        copyinstance = FiveCardStudPrivateState()
        if self.all_hand_cards is None:
            copyinstance.all_hand_cards = None
        else:
            copyinstance.all_hand_cards = [self.all_hand_cards[i].__deepcopy__() for i in xrange(len(self.all_hand_cards))]
        return copyinstance


class FiveCardStudPublicState(roomai.common.AbstractPublicState):
    """
    """
    first_hand_cards      = None
    second_hand_cards     = None
    third_hand_cards      = None
    fourth_hand_cards     = None
    fifth_hand_cards      = None

    is_quit               = None
    num_quit              = None
    is_raise              = None
    num_raise             = None
    is_needed_to_action   = None
    num_needed_to_action  = None

    # chips is array which contains the chips of all players
    chips = None

    # bets is array which contains the bets from all players
    bets = None

    upper_bet              = None
    floor_bet              = None
    max_bet_sofar          = None

    turn                   = None
    round                  = None
    num_players            = None

    previous_id            = None
    previous_action        = None
    previous_round         = None

    is_terminal            = None
    scores                 = None

    def __deepcopy__(self,memodict={}):
        """

        Args:
            memodict:

        Returns:

        """
        copyinstance = FiveCardStudPublicState()

        if self.first_hand_cards is None:
            copyinstance.first_hand_cards = None
        else:
            copyinstance.first_hand_cards = [self.first_hand_cards[i].__deepcopy__() for i in xrange(len(self.first_hand_cards))]

        if self.second_hand_cards is None:
            copyinstance.second_hand_cards = None
        else:
            copyinstance.second_hand_cards = [self.second_hand_cards[i].__deepcopy__() for i in xrange(len(self.second_hand_cards))]

        if self.third_hand_cards is None:
            copyinstance.third_hand_cards = None
        else:
            copyinstance.third_hand_cards = [self.third_hand_cards[i].__deepcopy__() for i in xrange(len(self.third_hand_cards))]

        if self.fourth_hand_cards is None:
            copyinstance.fourth_hand_cards = None
        else:
            copyinstance.fourth_hand_cards = [self.fourth_hand_cards[i].__deepcopy__() for i in xrange(len(self.fourth_hand_cards))]

        if self.fifth_hand_cards is None:
            copyinstance.fifth_hand_cards = None
        else:
            copyinstance.fifth_hand_cards = [self.fifth_hand_cards[i].__deepcopy__() for i in xrange(len(self.fifth_hand_cards))]

        copy.num_quit          = self.num_quit
        if self.is_quit is None:
            copyinstance.is_quit = None
        else:
            copyinstance.is_quit = [self.is_quit[i] for i in xrange(len(self.is_quit))]

        copyinstance.num_raise = self.num_raise
        if self.is_raise is None:
            copyinstance.is_raise = None
        else:
            copyinstance.is_raise  = [self.is_raise[i] for i in xrange(len(self.is_raise))]

        copyinstance.num_needed_to_action = self.num_needed_to_action
        if self.is_needed_to_action is None:
            copyinstance.is_needed_to_action = None
        else:
            copyinstance.is_needed_to_action = [self.is_needed_to_action[i] for i in xrange(len(self.is_needed_to_action))]

        # chips is array which contains the chips of all players
        if self.chips is None:
            copyinstance.chips = None
        else:
            copyinstance.chips = [self.chips[i] for i in xrange(len(self.chips))]


        # bets is array which contains the bets from all players
        if self.bets is None:
            copyinstance.bets = None
        else:
            copyinstance.bets = [self.bets[i] for i in xrange(len(self.bets))]

        copyinstance.upper_bet     = self.upper_bet
        copyinstance.floor_bet     = self.floor_bet
        copyinstance.max_bet_sofar = self.max_bet_sofar

        copyinstance.turn          = self.turn
        copyinstance.round         = self.round
        copyinstance.num_players   = self.num_players

        copyinstance.previous_id     = self.previous_id
        if self.previous_action is None:
            copyinstance.previous_action = None
        else:
            copyinstance.previous_action = self.previous_action.__deepcopy__()
        copyinstance.previous_round  = self.previous_round

        copyinstance.is_terminal = self.is_terminal

        if self.scores is None:
            copyinstance.scores = None
        else:
            copyinstance.scores = [self.scores[i] for i in xrange(len(self.scores))]

        return copyinstance


class FiveCardStudPersonState(roomai.common.AbstractPersonState):
    """
    """
    id                = None
    available_actions = None

    first_hand_card   = None
    second_hand_card  = None
    third_hand_card   = None
    fourth_hand_card  = None
    fifth_hand_card   = None


    def __deepcopy__(self, memodict={}, newinstance = None):
        """

        Args:
            memodict:
            newinstance:

        Returns:

        """
        copyinstance    = FiveCardStudPersonState()
        copyinstance.id = self.id

        if self.available_actions is not None:
            copyinstance.available_actions = dict()
            for key in self.available_actions:
                copyinstance.available_actions[key] = self.available_actions[key].__deepcopy__()
        else:
            copyinstance.available_actions = None

        if self.first_hand_card is not None:
            copyinstance.first_hand_card = self.first_hand_card.__deepcopy__()
        else:
            copyinstance.first_hand_card = None

        if self.second_hand_card is not None:
            copyinstance.second_hand_card = self.second_hand_card.__deepcopy__()
        else:
            copyinstance.second_hand_card = None

        if self.third_hand_card is not None:
            copyinstance.third_hand_card = self.third_hand_card.__deepcopy__()
        else:
            copyinstance.third_hand_card = None

        if self.fourth_hand_card is not None:
            copyinstance.fourth_hand_card = self.fourth_hand_card.__deepcopy__()
        else:
            copyinstance.fourth_hand_card = None

        if self.fifth_hand_card is not None:
            copyinstance.fifth_hand_card = self.fifth_hand_card.__deepcopy__()
        else:
            copyinstance.fifth_hand_card = None


        return copyinstance




#!/bin/python

import roomai.doudizhu

actions = roomai.doudizhu.DouDiZhuPokerEnv.available_actions_generate_all()
for action_key in actions:
    print action_key+"\t"+ \
          ",".join([str(s) for s in actions[action_key].masterCards])+"\t"+\
          ",".join([str(s) for s in actions[action_key].slaveCards])+"\t" +\
          actions[action_key].pattern[0]
action = roomai.doudizhu.DouDiZhuPokerAction([roomai.doudizhu.DouDiZhuActionElement.cheat],[])
print action.key + "\t" + \
      ",".join([str(s) for s in action.masterCards]) + "\t" + \
      ",".join([str(s) for s in action.slaveCards]) + "\t" + \
      action.pattern[0]

action = roomai.doudizhu.DouDiZhuPokerAction([roomai.doudizhu.DouDiZhuActionElement.bid],[])
print action.key + "\t" + \
      ",".join([str(s) for s in action.masterCards]) + "\t" + \
      ",".join([str(s) for s in action.slaveCards]) + "\t" + \
      action.pattern[0]




#!/bin/python
import roomai.common
from roomai.sevenking import SevenKingPublicState
from roomai.sevenking import SevenKingPrivateState
from roomai.sevenking import SevenKingPersonState
from roomai.sevenking import SevenKingAction
from roomai.sevenking import SevenKingPokerCard
from roomai.sevenking import AllSevenKingPatterns
from roomai.sevenking import AllSevenKingPokerCards
import random

import roomai.sevenking

logger = roomai.get_logger()

class SevenKingEnv(roomai.common.AbstractEnv):
    """
    """

    def init(self, params = dict()):
        """

        Args:
            params:

        Returns:

        """

        if "num_players" in params:
            self.num_players = params["num_players"]
        else:
            self.num_players = 3

        if "allcards" in params:
            allcards =  [c.__deepcopy__() for c in params["allcards"]]
        else:
            allcards =  [c.__deepcopy__() for c in AllSevenKingPokerCards.values()]
            random.shuffle(allcards)

        if "record_history" in params:
            self.record_history = params["record_history"]
        else:
            self.record_history = False




        self.public_state  = SevenKingPublicState()
        self.private_state = SevenKingPrivateState()
        self.person_states = [SevenKingPersonState() for i in range(self.num_players)]

        self.public_state_history  = []
        self.private_state_history = []
        self.person_states_history = []

        ## private_state
        self.private_state.keep_cards = allcards

        for i in range(self.num_players):
            self.person_states[i].hand_cards = []
            for j in range(5):
                c = self.private_state.keep_cards.pop()
                self.person_states[i].hand_cards.append(c)

        ## public_state
        self.public_state.turn,_          = self.choose_player_with_lowest_card()
        self.public_state.is_terminal     = False
        self.public_state.scores          = []
        self.public_state.previous_id     = None
        self.public_state.previous_action = None
        self.public_state.license_action  = SevenKingAction.lookup("")
        self.public_state.stage           = 0

        self.public_state.num_players     = self.num_players
        self.public_state.num_keep_cards  = len(self.private_state.keep_cards)
        self.public_state.num_hand_cards  = [len(person_state.hand_cards) for person_state in self.person_states]
        self.public_state.is_fold         = [False for i in range(self.public_state.num_players)]
        self.public_state.num_fold        = 0

        ## person_state
        for i in range(self.num_players):
            self.person_states[i].id         = i
            if i == self.public_state.turn:
                self.person_states[i].available_actions = SevenKingEnv.available_actions(self.public_state, self.person_states[i])

        self.__gen_history__()
        infos = self.__gen_infos__()
        return infos, self.public_state, self.person_states, self.private_state

    def forward(self, action):
        """

        Args:
            action:

        Returns:

        """
        pu   = self.public_state
        pr   = self.private_state
        pes  = self.person_states
        turn = pu.turn

        if SevenKingEnv.is_action_valid(action, pu, pes[turn]) == False:
            raise  ValueError("The (%s) is an invalid action " % (action.key))

        ## the action plays its role
        if action.pattern[0] == "p_0":
            pu.is_fold[turn]           = True
            pu.num_fold               += 1
            pes[turn].available_actions = dict()
        else:
            action_key_tmp  = dict([(c.key, None) for c in action.cards])
            cards_tmp       = pes[turn].hand_cards

            pes[turn].hand_cards = []
            for c in cards_tmp:
                if c.key not in action_key_tmp:
                    pes[turn].hand_cards.append(c)

            if pu.stage == 0:
                for i in range(5 - len(pes[turn].hand_cards)):
                    c = pr.keep_cards.pop()
                    pes[turn].hand_cards.append(c)
            elif pu.stage == 1:
                pu.num_hand_cards[turn] = len(pes[turn].hand_cards)

            pes[turn].available_actions = dict()

        pu.previous_id     = turn
        pu.previous_action = action.__deepcopy__()
        if action.pattern[0] != "p_0":
            pu.license_action = action



        #print (turn, "len_of_hand_card=",len(self.private_state.hand_cards[turn]), " len_of_keep_card=", len(self.private_state.keep_cards), " action = (%s)" %action.key,\
       #        " handcard1=%s"%(",".join([a.key for a in self.private_state.hand_cards[0]]))," handcard2=%s"%(",".join([a.key for a in self.private_state.hand_cards[1]])),\
         #      " num_fold =%d"%(self.public_state.num_fold),"fold=%s"%(",".join([str(s) for s in pu.is_fold])))
        ## termminal
        if self.public_state.stage == 1 and len(self.person_states[turn].hand_cards) == 0:
            pu.is_terminal    = True
            pu.scores         = self.compute_scores()
            new_turn          = None
            pu.turn           = new_turn
            pu.license_action = SevenKingAction.lookup("")

        ## stage 0 to 1
        elif len(self.private_state.keep_cards) < 5 and pu.stage == 0:
            new_turn, min_card              = self.choose_player_with_lowest_card()
            pu.turn                         = new_turn
            pu.num_fold                     = 0
            pu.is_fold                      = [False for i in range(pu.num_players)]
            pu.license_action               = SevenKingAction.lookup("")
            pes[new_turn].available_actions = SevenKingEnv.available_actions(pu, pes[new_turn])
            keys = pes[new_turn].available_actions.keys()
            for key in keys:
                if min_card.key not in key:
                    del pes[new_turn].available_actions[key]
            pu.stage                        = 1


        ## round next
        elif self.public_state.num_fold + 1 == pu.num_players:
            new_turn                        = self.choose_player_with_nofold()
            pu.turn                         = new_turn
            pu.num_fold                     = 0
            pu.is_fold                      = [False for i in range(pu.num_players)]
            pu.license_action               = SevenKingAction.lookup("")
            pes[new_turn].available_actions = SevenKingEnv.available_actions(pu, pes[new_turn])


        else:
            new_turn                        = (turn + 1) % pu.num_players
            pu.turn                         = new_turn
            pes[new_turn].available_actions = SevenKingEnv.available_actions(pu, pes[new_turn])



        self.__gen_history__()
        infos = self.__gen_infos__()
        return infos, self.public_state, self.person_states, self.private_state

    def compute_scores(self):
        """

        Returns:

        """
        scores                         = [-1 for i in range(self.num_players)]
        scores[self.public_state.turn] = self.num_players -1
        return scores

    def choose_player_with_nofold(self):
        """

        Returns:

        """
        for player_id in range(self.public_state.num_players):
            if self.public_state.is_fold[player_id]== False:
                return player_id



    def choose_player_with_lowest_card(self):
        """

        Returns:

        """
        min_card    = self.person_states[0].hand_cards[0]
        min_playerid = 0
        for playerid in range(self.num_players):
            for c in self.person_states[playerid].hand_cards:
                if SevenKingPokerCard.compare(min_card, c) > 0:
                    min_card     = c
                    min_playerid = playerid
        return min_playerid, min_card

    ######################## Utils function ###################
    @classmethod
    def compete(cls, env, players):
        """

        Args:
            env:
            players:

        Returns:

        """

        num_players = len(players)
        infos, public_state, person_states, private_state = env.init({"num_players":num_players})
        for i in range(env.num_players):
            players[i].receive_info(infos[i])

        while public_state.is_terminal == False:
            turn   = public_state.turn
            action = players[turn].take_action()
            infos, public_state, person_states, private_state = env.forward(action)
            for i in range(env.num_players):
                players[i].receive_info(infos[i])

        return public_state.scores




    @classmethod
    def is_action_valid(self, action, public_state, person_state):
        """

        Args:
            action:
            public_state:
            person_state:

        Returns:

        """
        license_action = public_state.license_action
        if license_action is None:
            license_action = SevenKingAction.lookup("")

        if action.pattern[0] == "p_0":
            if license_action.pattern[0] != "p_0":   return True
            elif license_action.pattern[0] == "p_0":
                logger.error("The p_0 type action is invalid in the begining of the game or after the previous player took the p_0 type action ")
                return False

        ### is action from hand_cards
        hand_keys = dict()
        for c in person_state.hand_cards:
            key = c.key
            if c not in hand_keys:  hand_keys[key] = 1
            else:  hand_keys[c] += 1
        action_keys = dict()
        for c in action.cards:
            key = c.key
            if c not in action_keys: action_keys[key] = 1
            else:action_keys[key] +=1

        for k in action_keys:
            if k not in hand_keys or hand_keys[k] < action_keys[k]:
                return False

        ## pattern
        if license_action.pattern[0] != "p_0" and license_action.pattern[0] != action.pattern[0]:
            return False
        if license_action.pattern[0] == "p_0" and action.pattern == "p_0":
            return False


        if license_action.pattern[0] != "p_0":
            max_action_card = action.cards[0]
            for c in action.cards:
                if SevenKingPokerCard.compare(max_action_card,c) < 0:
                    max_action_card = c
            max_previous_card = license_action.cards[0]
            for c in license_action.cards:
                if SevenKingPokerCard.compare(max_previous_card,c) < 0:
                    max_previous_card = c
            if SevenKingPokerCard.compare(max_action_card, max_previous_card) < 0 :
                return False

        return True



    ########################### about gen_available_actions ########################
    @classmethod
    def __gen_available_actions_with_pattern(cls, hand_card, pattern):
        """

        Args:
            hand_card:
            pattern:

        Returns:

        """
        res = []

        if len(hand_card) < pattern[1]:
            return res
        if pattern[0] == "p_0":
            return res

        point2cards = dict()
        for c in hand_card:
            point = c.get_point_rank()
            if point not in point2cards:
                point2cards[point] = []
            point2cards[point].append(c.__deepcopy__())
        for p in point2cards:
            point2cards[p].sort(cmp = SevenKingPokerCard.compare)

        if pattern[0] == "p_1":
            for c in hand_card:
                res.append(SevenKingAction.lookup(c.key))

        elif pattern[0] == "p_2":
            for p in point2cards:
                len1 = len(point2cards[p])
                if len1 == 2:
                    str = "%s,%s"%(point2cards[p][0].key,point2cards[p][1].key)
                    res.append(SevenKingAction.lookup(str))
                if len1 == 3:
                    str = "%s,%s"%(point2cards[p][0].key,point2cards[p][1].key)
                    res.append(SevenKingAction.lookup(str))
                    str = "%s,%s"%(point2cards[p][0].key,point2cards[p][2].key)
                    res.append(SevenKingAction.lookup(str))
                    str = "%s,%s"%(point2cards[p][1].key,point2cards[p][2].key)
                    res.append(SevenKingAction.lookup(str))
                if len1 == 4:
                    str = "%s,%s" % (point2cards[p][0].key, point2cards[p][1].key)
                    res.append(SevenKingAction.lookup(str))
                    str = "%s,%s" % (point2cards[p][0].key, point2cards[p][2].key)
                    res.append(SevenKingAction.lookup(str))
                    str = "%s,%s" % (point2cards[p][0].key, point2cards[p][3].key)
                    res.append(SevenKingAction.lookup(str))
                    str = "%s,%s" % (point2cards[p][1].key, point2cards[p][2].key)
                    res.append(SevenKingAction.lookup(str))
                    str = "%s,%s" % (point2cards[p][1].key, point2cards[p][3].key)
                    res.append(SevenKingAction.lookup(str))
                    str = "%s,%s" % (point2cards[p][2].key, point2cards[p][3].key)
                    res.append(SevenKingAction.lookup(str))


        elif pattern[0] == "p_3":
            for p in point2cards:
                len1 = len(point2cards[p])
                if len1 == 3:
                    str = "%s,%s,%s" % (point2cards[p][0].key, point2cards[p][1].key, point2cards[p][2].key)
                    res.append(SevenKingAction.lookup(str))
                if len1 == 4:
                    str = "%s,%s,%s" % (point2cards[p][0].key, point2cards[p][1].key, point2cards[p][2].key)
                    res.append(SevenKingAction.lookup(str))
                    str = "%s,%s,%s" % (point2cards[p][0].key, point2cards[p][1].key, point2cards[p][3].key)
                    res.append(SevenKingAction.lookup(str))
                    str = "%s,%s,%s" % (point2cards[p][0].key, point2cards[p][2].key, point2cards[p][3].key)
                    res.append(SevenKingAction.lookup(str))
                    str = "%s,%s,%s" % (point2cards[p][1].key, point2cards[p][2].key, point2cards[p][3].key)
                    res.append(SevenKingAction.lookup(str))

        elif pattern[0] == "p_4":
            for p in point2cards:
                if len(point2cards[p]) >= 4:
                    str = "%s,%s,%s,%s"%(
                        point2cards[p][0].key,
                        point2cards[p][1].key,
                        point2cards[p][2].key,
                        point2cards[p][3].key
                    )
                    res.append(SevenKingAction.lookup(str))

        else:
            raise ValueError("The %s pattern is invalid"%(pattern[0]))

        return res

    @classmethod
    def available_actions(cls, public_state, person_state):
        """

        Args:
            public_state:
            person_state:

        Returns:

        """
        available_actions = dict()

        license_action = public_state.license_action
        if license_action is None:
            license_action = SevenKingAction("")
        hand_cards = person_state.hand_cards


        if license_action.pattern[0] == "p_0":
            for pattern in AllSevenKingPatterns.values():
                if pattern[0] == "p_0":continue
                actions = cls.__gen_available_actions_with_pattern(hand_cards, pattern)
                for action in actions:
                    if cls.is_action_valid(action, public_state, person_state) == True:
                        available_actions[action.key] = action
        else:
            actions = cls.__gen_available_actions_with_pattern(hand_cards, license_action.pattern)
            for action in actions:
                if cls.is_action_valid(action, public_state, person_state) == True:
                    available_actions[action.key] = action
            available_actions[""] = SevenKingAction.lookup("")

        return available_actions




#!/bin/python
import roomai.common
import random
from roomai.fivecardstud import FiveCardStudEnv

class FiveCardStudRandomPlayer(roomai.common.AbstractPlayer):
    """
    """

    public_state = None
    person_state = None
    def receive_info(self, info):
        """

        Args:
            info:
        """
        self.public_state = info.public_state
        self.person_state = info.person_state

    def take_action(self):
        """

        Returns:

        """
        actions = FiveCardStudEnv.available_actions(self.public_state, self.person_state).values()
        idx     = int(random.random() * len(actions))
        return actions[idx]

    def reset(self):
        """

        """
        pass



#!/bin/python
import random
import math
import copy
import roomai.common
import roomai.kuhn.KuhnPokerUtils

class KuhnPokerEnv(roomai.common.AbstractEnv):
    """
    """

    #@override
    def init(self, params=dict()):
        """

        Args:
            params:

        Returns:

        """

        if "record_history" in params:
            self.record_history = params["record_history"]
        else:
            self.record_history = False

        if "start_turn" in params:
            self.start_turn = params["start_turn"]
        else:
            self.start_turn = int(random.random() * 2)

        self.available_action = dict()
        self.available_action[roomai.kuhn.KuhnPokerUtils.KuhnPokerAction("check").key] = roomai.kuhn.KuhnPokerAction.lookup("check")
        self.available_action[roomai.kuhn.KuhnPokerUtils.KuhnPokerAction("bet").key]   = roomai.kuhn.KuhnPokerAction.lookup("bet")

        self.private_state = roomai.kuhn.KuhnPokerUtils.KuhnPokerPrivateState()
        self.public_state  = roomai.kuhn.KuhnPokerUtils.KuhnPokerPublicState()
        self.person_states = [roomai.kuhn.KuhnPokerUtils.KuhnPokerPersonState() for i in xrange(2)]

        self.public_state_history  = []
        self.private_state_history = []
        self.person_states_history = []

        card0 = math.floor(random.random() * 3)
        card1 = math.floor(random.random() * 3)
        while card0 == card1:
            card0 = math.floor(random.random() * 3)
        self.private_state.hand_cards = [card0, card1]

        self.public_state.turn          = self.start_turn
        self.public_state.first         = self.public_state.turn
        self.public_state.epoch         = 0
        self.public_state.action_list   = []
        self.public_state.is_terminal   = False
        self.public_state.scores        = None
        self.person_states[0].id = 0
        self.person_states[0].card      = card0
        self.person_states[1].id        = 1
        self.person_states[1].card      = card1

        self.person_states[self.public_state.turn].available_actions = self.available_action

        self.__gen_history__()
        infos = self.__gen_infos__()

        
        return  infos, self.public_state, self.person_states, self.private_state

    #@override
    def forward(self, action):
        """

        Args:
            action:

        Returns:

        """
        self.person_states[self.public_state.turn].available_actions = dict()
        self.public_state.epoch                                     += 1
        self.public_state.turn                                       = (self.public_state.turn+1)%2
        self.public_state.action_list.append(action.key)

        if self.public_state.epoch == 1:
            self.public_state.is_terminal = False
            self.public_state.scores      = []
            self.person_states[self.public_state.turn].available_actions = self.available_action

            self.__gen_history__()
            infos = self.__gen_infos__()
            return infos, self.public_state, self.person_states, self.private_state

        elif self.public_state.epoch == 2:
            scores = self.evaluteTwo()
            if scores is not None:
                self.public_state.is_terminal = True
                self.public_state.scores      = scores

                self.__gen_history__()
                infos = self.__gen_infos__()
                return infos,self.public_state, self.person_states, self.private_state
            else:
                self.public_state.is_terminal = False
                self.public_state.scores      = []
                self.person_states[self.public_state.turn].available_actions = self.available_action

                self.__gen_history__()
                infos                         = self.__gen_infos__()
                return infos,self.public_state, self.person_states, self.private_state

        elif self.public_state.epoch == 3:
            self.public_state.is_terminal = True
            self.public_state.scores      = self.evaluteThree()

            self.__gen_history__()
            infos                         = self.__gen_infos__()
            return infos,self.public_state, self.person_states, self.private_state

        else:
            raise Exception("KuhnPoker has 3 turns at most")


    #@Overide
    @classmethod
    def compete(cls, env, players):
        """

        Args:
            env:
            players:

        Returns:

        """

        infos, public_state, person_state, private_state = env.init()
        for i in xrange(len(players)):
            players[i].receive_info(infos[i])

        while public_state.is_terminal == False:
            turn = infos[-1].public_state.turn
            action = players[turn].take_action()
            infos,public_state, person_state, private_state = env.forward(action)
            for i in xrange(len(players)):
                players[i].receive_info(infos[i])

        return public_state.scores



    def WhoHasHigherCard(self):
        """

        Returns:

        """
        hand_cards = self.private_state.hand_cards
        if hand_cards[0] > hand_cards[1]:
            return 0
        else:
            return 1

    def evaluteTwo(self):
        """

        Returns:

        """
        win    = self.WhoHasHigherCard()
        first  = self.public_state.first
        scores = [0, 0];
        actions = self.public_state.action_list

        if actions[0] == "check" and \
           actions[1] == "bet":
            return None
        
        if actions[0] == actions[1] and \
           actions[0] == "check":
            scores[win]   = 1;
            scores[1-win] = -1
            return scores;

        if actions[0] == "bet" and \
           actions[1] == "check":
            scores[first]   = 1;
            scores[1-first] = -1
            return scores;

        if actions[0] == actions[1] and \
           actions[0] == "bet":
            scores[win]   = 2
            scores[1-win] = -2
            return scores;


    def evaluteThree(self):
        """

        Returns:

        """
        first   = self.public_state.first 
        win     = self.WhoHasHigherCard()
        scores  = [0, 0]

        if self.public_state.action_list[2] == "check":
            scores[1 - first] = 1;
            scores[first]     = -1
        else:
            scores[win]   = 2;
            scores[1-win] = -2
        return scores;
       



#!/bin/python
import roomai.common
import copy


class KuhnPokerAction(roomai.common.AbstractAction):
    """
    """
    bet   = 0;
    check = 1;
    def __init__(self, key):
        """

        Args:
            key:
        """
        self.__key = ""
        if key == "bet": self.__key = KuhnPokerAction.bet
        elif key == "check":self.__key = KuhnPokerAction.check
        else:
            raise KeyError("%s is invalid key for Kuhn DouDiZhuPokerAction"%(key))

    @property
    def key(self):
        """

        Returns:

        """
        if self.__key == KuhnPokerAction.bet:
            return "bet"
        else:
            return "check"

    @classmethod
    def lookup(cls, key):
        """

        Args:
            key:

        Returns:

        """
        return AllKuhnActions[key]

    def __deepcopy__(self, memodict={}):
        """

        Args:
            memodict:

        Returns:

        """
        copy = KuhnPokerAction(self.key)
        return copy

AllKuhnActions = {"bet":KuhnPokerAction("bet"),"check":KuhnPokerAction("check")}

class KuhnPokerPublicState(roomai.common.AbstractPublicState):
    """
    """
    def __init__(self):
        """

        """
        self.turn                       = 0
        self.first                      = 0
        self.epoch                      = 0
        self.action_list                = []


    def __deepcopy__(self, memodict={}):
        """

        Args:
            memodict:

        Returns:

        """
        copy = KuhnPokerPublicState()
        copy.turn  = self.turn
        copy.first = self.first
        copy.epoch = self.epoch
        for a in self.action_list:
            copy.action_list.append(a)
        return copy

class KuhnPokerPrivateState(roomai.common.AbstractPrivateState):
    """
    """
    def __init__(self):
        """

        """
        self.hand_cards = []

    def __deepcopy__(self, memodict={}):
        """

        Args:
            memodict:

        Returns:

        """
        copy            = KuhnPokerPrivateState()
        copy.hand_cards = [card for card in self.hand_cards]
        return copy

class KuhnPokerPersonState(roomai.common.AbstractPersonState):
    """
    """
    def __init__(self):
        """

        """
        self.available_actions  = dict()
        self.id                 = 1
        self.card               = 0

    def __deepcopy__(self, memodict={}, newinstance = None):
        """

        Args:
            memodict:
            newinstance:

        Returns:

        """
        copy                   = KuhnPokerPersonState()
        copy.id                = self.id
        copy.card              = self.card
        copy.available_actions = dict()

        if self.available_actions is None:
            copy.available_actions = None
        else:
            for action_key in self.available_actions:
                copy.available_actions[action_key] = self.available_actions[action_key].__deepcopy__()
        return copy



#!/bin/python

from roomai.doudizhu.DouDiZhuPokerAction import AllActions
from roomai.doudizhu.DouDiZhuPokerAction import AllPatterns
from roomai.doudizhu.DouDiZhuPokerAction import DouDiZhuPokerAction
from roomai.doudizhu.DouDiZhuPokerAction import DouDiZhuActionElement
from roomai.doudizhu.DouDiZhuPokerEnv    import *
from roomai.doudizhu.DouDiZhuPokerPlayer import *




#!/bin/python
from FiveCardStudAction import FiveCardStudAction
from FiveCardStudInfo   import FiveCardStudPublicState,FiveCardStudPersonState,FiveCardStudPrivateState
from FiveCardStudUtils  import FiveCardStudPokerCard,FiveCardStudAllCardsPattern
from FiveCardStudEnv    import FiveCardStudEnv
from FiveCardStudPlayer import FiveCardStudRandomPlayer


#!/bin/python
#coding=utf8

import roomai
logger = roomai.get_logger()

######################################################################### Basic Concepts #####################################################
class AbstractPublicState(object):
    """
    """
    def __init__(self):
        """

        """
        self.turn            = None
        self.previous_id     = None
        self.previous_action = None

        self.is_terminal     = False
        self.scores          = None

    def __deepcopy__(self, memodict={}, newinstance = None):
        """

        Args:
            memodict:
            newinstance:

        Returns:

        """
        if newinstance is  None:
            newinstance = AbstractPublicState()
        newinstance.turn             = self.turn
        newinstance.previous_id      = self.previous_id
        if self.previous_action is not None:
            newinstance.previous_action  = self.previous_action.__deepcopy__()
        else:
            newinstance.previous_action  = None
        newinstance.is_terminal      = self.is_terminal
        newinstance.scores           = [score for score in self.scores]
        return newinstance


class AbstractPrivateState(object):
    """
    """
    def __deepcopy__(self, newinstance = None, memodict={}):
        """

        Args:
            newinstance:
            memodict:

        Returns:

        """
        if newinstance is  None:
            return AbstractPrivateState()
        else:
            return newinstance


class AbstractPersonState(object):
    """
    """
    def __init__(self):
        """

        """
        self.id                = 0
        self.available_actions = dict()
    def __deepcopy__(self, memodict={}, newinstance = None):
        """

        Args:
            memodict:
            newinstance:

        Returns:

        """
        #print "enter AbstractPersonState __deepcopy__"
        if newinstance is  None:
            newinstance = AbstractPersonState()
        newinstance.id                = self.id
        newinstance.available_actions = dict()
        for k in self.available_actions:
            #print "fuck"
            newinstance.available_actions[k] = self.available_actions[k].__deepcopy__()
        return newinstance

class Info(object):
    """
    """
    def __init__(self):
        """

        """
        self.public_state       = AbstractPublicState()
        self.person_state       = AbstractPersonState()
    def __deepcopy__(self, memodict={}, newinstance = None):
        """

        Args:
            memodict:
            newinstance:

        Returns:

        """
        if newinstance is None:
            newinstance = Info()
        newinstance.public_state = self.public_state.__deepcopy__()
        newinstance.public_state = self.person_state.__deepcopy__()
        return newinstance

class AbstractAction(object):
    """
    """
    def __init__(self, key):
        """

        Args:
            key:
        """
        self.__key = key

    @property
    def key(self):
        """

        Returns:

        """
        return self.__key

    @classmethod
    def lookup(self, key):
        """

        Args:
            key:
        """
        raise NotImplementedError("Not implemented")

    def __deepcopy__(self, memodict={}, newinstance = None):
        """

        Args:
            memodict:
            newinstance:

        Returns:

        """
        if newinstance is None:
            newinstance = AbstractAction()
        newinstance.__key = self.__key
        return newinstance




class AbstractPlayer(object):
    """
    """
    def receive_info(self, info):
        """
        :param:
            info: the information produced by a game environments

        Raises:
                NotImplementedError: An error occurred when we doesn't implement this function
        """
        raise NotImplementedError("The receiveInfo function hasn't been implemented") 

    def take_action(self):
        """
        Returns:
            A DouDiZhuPokerAction produced by this player
        """
        raise NotImplementedError("The takeAction function hasn't been implemented") 

    def reset(self):
        """

        """
        raise NotImplementedError("The reset function hasn't been implemented")




class AbstractEnv(object):
    """
    """
    public_state_history = []
    private_state_history = []
    person_states_history = []

    def __init__(self, params=dict()):
        """

        Args:
            params:
        """
        self.public_state   = AbstractPublicState()
        self.private_state  = AbstractPrivateState()
        self.person_states  = [AbstractPrivateState()]
        self.record_history = False


    def __gen_infos__(self):
        """

        Returns:

        """
        num_players = len(self.person_states)
        infos = [Info() for i in xrange(num_players)]
        for i in xrange(num_players):
            infos[i].person_state = self.person_states[i].__deepcopy__()
            infos[i].public_state = self.public_state.__deepcopy__()

        return infos

    def __gen_history__(self):
        """

        Returns:

        """
        if self.record_history == False:
            return

        self.public_state_history.append(self.public_state.__deepcopy__())
        self.private_state_history.append(self.private_state.__deepcopy__())
        self.person_states_history.append([person_state.__deepcopy__() for person_state in self.person_states])

    def init(self, params =dict()):
        """
        Args:
            chance_action: 

        Returns:
            infos, public_state, person_states, private_state, other_chance_actions
        """
        raise ("The init function hasn't been implemented")

    def forward(self, action, params = None):
        """
        :param action, chance_action

        Returns:
            infos, public_state, person_states, private_state, other_chance_actions
        """
        raise NotImplementedError("The forward hasn't been implemented")


    def backward(self):
        """
        The game goes back to the previous states

        Returns:
            infos, public_state, person_states, private_state

        :ValueError: if Env has reached the initializaiton state, and we call this backward function, we will get ValueError.
        """

        if self.record_history == False:
            raise ValueError("Env can't backward when record_history = False. params = {\"record_history\":true} and env.init(params) will set the record_history to be true")

        if len(self.public_state_history) == 1:
            raise ValueError("Env has reached the initialization state and can't go back further. ")
        self.public_state_history.pop()
        self.private_state_history.pop()
        self.person_states_history.pop()

        p = len(self.public_state_history) - 1
        self.public_state  = self.public_state_history[p].__deepcopy__()
        self.private_state = self.private_state_history[p].__deepcopy__()
        self.person_states = [person_state.__deepcopy__() for person_state in self.person_states_history[p]]

        infos  = self.__gen_infos__()
        return infos, self.public_state, self.person_states, self.private_state

    ### provide some util functions
    @classmethod
    def compete(cls, env, players):
        """
        Args:
            env: 
            players: 

        Returns:
            [score_for_player0, score_for_player1,...]
        """
        raise NotImplementedError("The round function hasn't been implemented")

    @classmethod
    def is_action_valid(cls, action, public_state, person_state):
        """
        :param action

        Args:
            public_state: 
            person_state: 

        Returns:
            is  the action valid
        """
        raise  NotImplementedError("The is_action_valid function hasn't been implemented")

    @classmethod
    def available_actions(self, public_state, person_state):
        """
        Args:
            public_state: 
            person_state: 

        Returns:
            all available_actions
        """
        raise NotImplementedError("The available_actions function hasn't been implemented")

############################################################### Some Utils ############################################################################

point_str_to_rank  = {'2':0, '3': 1, '4': 2, '5': 3, '6': 4, '7': 5, '8':6, '9':7, 'T':8, 'J':9, 'Q':10, 'K':11, 'A':12, 'r':13, 'R':14}
point_rank_to_str  = {0: '2', 1: '3', 2: '4', 3: '5', 4: '6', 5: '7', 6: '8', 7: '9', 8: 'T', 9: 'J', 10: 'Q', 11: 'K', 12: 'A', 13: 'r', 14: 'R'}
suit_str_to_rank   = {'Spade':0, 'Heart':1, 'Diamond':2, 'Club':3,  'ForKing':4}
suit_rank_to_str   = {0:'Spade', 1: 'Heart', 2: 'Diamond', 3:'Club', 4:'ForKing'}


class PokerCard(object):
    """
    """
    def __init__(self, point, suit = None):
        """

        Args:
            point:
            suit:
        """
        point1 = 0
        suit1  = 0
        if suit is None:
            kv = point.split("_")
            point1 = point_str_to_rank[kv[0]]
            suit1  = suit_str_to_rank[kv[1]]
        else:
            point1 = point
            if isinstance(point, str):
                point1 = point_str_to_rank[point]
            suit1  = suit
            if isinstance(suit, str):
                suit1 = suit_str_to_rank[suit]

        self.__point_str = point_rank_to_str[point1]
        self.__suit_str  = suit_rank_to_str[suit1]
        self.__key = "%s_%s" % (self.point_str, self.suit_str)

    @property
    def point_str(self):
        """

        Returns:

        """
        return self.__point_str

    @property
    def suit_str(self):
        """

        Returns:

        """
        return self.__suit_str

    @property
    def point_rank(self):
        """

        Returns:

        """
        return point_str_to_rank[self.__point_str]
    @property
    def suit_rank(self):
        """

        Returns:

        """
        return

    @property
    def key(self):
        """

        Returns:

        """
        return self.__key

    @classmethod
    def lookup(cls, key):
        """

        Args:
            key:

        Returns:

        """
        return AllPokerCards[key]

    def get_point_rank(self):
        """

        Returns:

        """
        return point_str_to_rank[self.point_str]

    def get_suit_rank(self):
        """

        Returns:

        """
        return suit_str_to_rank[self.suit_str]

    @classmethod
    def compare(cls, pokercard1, pokercard2):
        """

        Args:
            pokercard1:
            pokercard2:

        Returns:

        """
        pr1 = pokercard1.get_point_rank()
        pr2 = pokercard2.get_point_rank()

        if pr1 != pr2:
            return pr1 - pr2
        else:
            return pokercard1.get_suit_rank() - pokercard2.get_suit_rank()

    def __deepcopy__(self,  memodict={}, newinstance = None):
        """

        Args:
            memodict:
            newinstance:

        Returns:

        """
        if newinstance is None:
            newinstance = AllPokerCards[self.key]
        return newinstance
    def lookup(self, key):
        """

        Args:
            key:
        """
        AllPokerCards[key]

AllPokerCards = dict()
for point_str in point_str_to_rank:
    if point_str != 'r' and point_str != "R":
        for suit_str in suit_str_to_rank:
            if suit_str != "ForKing":
                AllPokerCards["%s_%s"%(point_str,suit_str)] = PokerCard("%s_%s"%(point_str,suit_str))
AllPokerCards["r_ForKing"] = (PokerCard("r_ForKing"))
AllPokerCards["R_ForKing"] = (PokerCard("R_ForKing"))



#!/bin/python
#coding:utf-8
import roomai.common
import copy
import logging
import random
import sys


from FiveCardStudUtils  import FiveCardStudPokerCard
from FiveCardStudInfo   import FiveCardStudPublicState
from FiveCardStudInfo   import FiveCardStudPersonState
from FiveCardStudInfo   import FiveCardStudPrivateState
from FiveCardStudAction import FiveCardStudAction

class FiveCardStudEnv(roomai.common.AbstractEnv):
    """
    """


    #@override
    def init(self, params = dict()):
        """

        Args:
            params:

        Returns:

        """
        self.logger         = roomai.get_logger()

        if "num_players" in params:
            self.num_players = params["num_players"]
        else:
            self.num_players = 3

        if "chips" in params:
            self.chips       = params["chips"]
        else:
            self.chips       = [500 for i in range(self.num_players)]

        if "floor_bet" in params:
            self.floor_bet   = params["floor_bet"]
        else:
            self.floor_bet   = 10

        if "allcards" in params:
            self.allcards    = params["allcards"]
        else:
            self.allcards = []
            for i in xrange(13):
                for j in xrange(4):
                    self.allcards.append(FiveCardStudPokerCard(i, j))
            random.shuffle(self.allcards)

        if "record_history" in params:
            self.record_history = params["record_history"]
        else:
            self.record_history = False


        FiveCardStudEnv.check_initialization_configuratioin(self)

        self.public_state   = FiveCardStudPublicState()
        self.private_state  = FiveCardStudPrivateState()
        self.person_states  = [FiveCardStudPersonState() for i in xrange(self.num_players)]

        self.public_state_history  = []
        self.private_state_history = []
        self.person_states_history = []



        ## private_state
        self.private_state.all_hand_cards      = self.allcards[0: 5 * self.num_players]

        ## public_state
        self.public_state.num_players          = self.num_players
        self.public_state.chips                = self.chips
        self.public_state.second_hand_cards    = self.private_state.all_hand_cards[1*self.num_players:  2 * self.num_players]
        self.public_state.floor_bet            = self.floor_bet
        self.public_state.upper_bet            = min(self.public_state.chips)
        #print "public_state.upper_bet", self.public_state.upper_bet,"chips", self.public_state.chips

        self.public_state.bets                 = [self.public_state.floor_bet for i in xrange(self.num_players)]
        self.public_state.chips                = [self.public_state.chips[i] - self.public_state.floor_bet for i in xrange(self.num_players)]
        self.public_state.max_bet_sofar        = self.public_state.floor_bet
        self.public_state.is_quit              = [False for i in xrange(self.num_players)]
        self.public_state.num_quit             = 0
        self.public_state.is_needed_to_action  = [True for i in xrange(self.num_players)]
        self.public_state.num_needed_to_action = self.num_players
        self.public_state.is_raise             = [False for i in xrange(self.num_players)]
        self.public_state.num_raise            = 0

        self.public_state.round                = 1
        self.public_state.turn                 = FiveCardStudEnv.choose_player_at_begining_of_round(self.public_state)
        self.public_state.is_terminal          = False
        self.public_state.scores               = None

        ## person_state
        for i in xrange(self.num_players):
            self.person_states[i].id = i
            self.person_states[i].first_hand_card  = self.private_state.all_hand_cards[i]
            self.person_states[i].second_hand_card = self.private_state.all_hand_cards[self.num_players+i]
        turn = self.public_state.turn
        self.person_states[turn].available_actions = FiveCardStudEnv.available_actions(self.public_state, self.person_states[turn])

        self.__gen_history__()
        infos = self.__gen_infos__()

        return infos, self.public_state, self.person_states, self.private_state


    #@override
    def forward(self, action):
        """
        Args:
            action: 

        Raises:
            None: throw ValueError when the action is invalid at this time
        """
        turn = self.public_state.turn
        if not FiveCardStudEnv.is_action_valid(action,self.public_state, self.person_states[turn]):
            self.logger.critical("action=%s is invalid" % (action.key()))
            raise ValueError("action=%s is invalid" % (action.key()))

        pu = self.public_state
        pe = self.person_states
        pr = self.private_state

        if action.option == FiveCardStudAction.Fold:
            self.action_fold(action)
        elif action.option == FiveCardStudAction.Check:
            self.action_check(action)
        elif action.option == FiveCardStudAction.Call:
            self.action_call(action)
        elif action.option == FiveCardStudAction.Raise:
            self.action_raise(action)
        elif action.option == FiveCardStudAction.Showhand:
            self.action_showhand(action)
        elif action.option == FiveCardStudAction.Bet:
            self.action_bet(action)
        else:
            raise Exception("action.option(%s) not in [Fold, Check_, Call, Raise, Showhand, Bet]" % (action.option))
        pu.previous_id     = pu.turn
        pu.previous_action = action
        pu.previous_round  = pu.round

        # computing_score
        if FiveCardStudEnv.is_compute_scores(self.public_state):
            num_players          = pu.num_players
            pu.hand_cards        = []
            pu.first_hand_cards  = pr.all_hand_cards[0:                1 * num_players]
            pu.second_hand_cards = pr.all_hand_cards[1 * num_players:  2 * num_players]
            pu.third_hand_cards  = pr.all_hand_cards[2 * num_players:  3 * num_players]
            pu.fourth_hand_cards = pr.all_hand_cards[3 * num_players:  4 * num_players]
            pu.fifth_hand_cards  = pr.all_hand_cards[4 * num_players:  5 * num_players]
            pu.round             = 4


            pu.is_terminal = True
            pu.scores      = self.compute_scores(pu)

            for i in xrange(num_players):
                pu.chips[i] += pu.bets[i] + pu.scores[i]

            for i in xrange(num_players):
                pe[i].first_hand_card  = pr.all_hand_cards[0 * num_players + i]
                pe[i].second_hand_card = pr.all_hand_cards[1 * num_players + i]
                pe[i].third_hand_card  = pr.all_hand_cards[2 * num_players + i]
                pe[i].fourth_hand_card = pr.all_hand_cards[3 * num_players + i]
                pe[i].fifth_hand_card  = pr.all_hand_cards[4 * num_players + i]

            pu.turn                              = None
            pe[pu.previous_id].available_actions = None

        # enter into the next stage
        elif FiveCardStudEnv.is_nextround(self.public_state):
            num_players = self.public_state.num_players
            add_cards   = []
            if pu.round == 1:
                pu.third_hand_cards        = pr.all_hand_cards[2 * num_players:  3 * num_players]
                for i in xrange(num_players):
                    pe[i].third_hand_card  = pr.all_hand_cards[2 * num_players + i]
            if pu.round == 2:
                pu.fourth_hand_cards       = pr.all_hand_cards[3 * num_players:  4 * num_players]
                for i in xrange(num_players):
                    pe[i].fourth_hand_card = pr.all_hand_cards[3 * num_players + i]
            if pu.round == 3:
                pu.fifth_hand_cards        = pr.all_hand_cards[4 * num_players:  5 * num_players]
                for i in xrange(num_players):
                    pe[i].fifth_hand_card  = pr.all_hand_cards[4 * num_players + i]


            pu.round                = pu.round + 1
            pu.turn                 = FiveCardStudEnv.choose_player_at_begining_of_round(pu)

            pu.num_needed_to_action = 0
            for i in range(self.num_players):
                if pu.is_quit[i] == False and pu.bets[i] < pu.upper_bet:
                    pu.is_needed_to_action[i] = True
                    pu.num_needed_to_action  += 1
                pu.is_raise[i]            = False
            pu.num_raise = 0

            pe[pu.previous_id].available_actions = None
            pe[pu.turn].available_actions        = FiveCardStudEnv.available_actions(pu, pe[pu.turn])

        else:
            pu.turn                              = self.next_player(pu)
            pe[pu.previous_id].available_actions = None
            pe[pu.turn].available_actions        = FiveCardStudEnv.available_actions(pu, pe[pu.turn])


        self.__gen_history__()
        infos  = self.__gen_infos__()


        return infos, self.public_state, self.person_states, self.private_state


    #@override
    @classmethod
    def compete(cls, env, players):
        """

        Args:
            env:
            players:

        Returns:

        """
        total_scores   = [0 for i in xrange(len(players))]
        total_count    = 1000

        for count in xrange(total_count):
            chips       = [(100 +int(random.random()*500)) for i in xrange(len(players))]
            num_players = len(players)
            floor_bet   = 10
            infos, public, persons, private = env.init({"num_players":num_players,"chips":chips, "floor_bet":10})
            for i in xrange(len(players)):
                players[i].receive_info(infos[i])

            while public.is_terminal == False:
                turn = public.turn
                action = players[turn].take_action()
                infos, public, persons, private = env.forward(action)
                for i in xrange(len(players)):
                    players[i].receive_info(infos[i])

            for i in xrange(len(players)):
                players[i].reset()
                total_scores[i] += public.scores[i]


            if (count + 1)%500 == 0:
                tmp_scores = [0 for i in xrange(len(total_scores))]
                for i in xrange(len(total_scores)):
                    tmp_scores[i] = total_scores[i] / (count+1)
                roomai.get_logger().info("FiveCardStud completes %d competitions, scores=%s"%(count+1, ",".join([str(i) for i in tmp_scores])))

        for i in xrange(len(players)):
            total_scores[i]   /= total_count * 1.0

        return total_scores;


    def action_fold(self, action):
        """

        Args:
            action:
        """
        pu = self.public_state
        pu.is_quit[pu.turn] = True
        pu.num_quit        += 1

        pu.is_needed_to_action[pu.turn] = False
        pu.num_needed_to_action        -= 1

    def action_check(self, action):
        """

        Args:
            action:
        """
        pu = self.public_state
        pu.is_needed_to_action[pu.turn] = False
        pu.num_needed_to_action        -= 1

    def action_call(self, action):
        """

        Args:
            action:
        """
        pu = self.public_state
        pu.chips[pu.turn]               -= action.price
        pu.bets[pu.turn]                += action.price
        pu.is_needed_to_action[pu.turn] = False
        pu.num_needed_to_action        -= 1


    def action_bet(self, action):
        """

        Args:
            action:
        """
        pu = self.public_state

        pu.chips[pu.turn] -= action.price
        pu.bets[pu.turn]  += action.price
        pu.max_bet_sofar   = pu.bets[pu.turn]

        pu.is_needed_to_action[pu.turn] = False
        pu.num_needed_to_action        -= 1
        p = (pu.turn + 1)%pu.num_players
        while p != pu.turn:
            if  pu.is_quit[p] == False and pu.is_needed_to_action[p] == False and pu.bets[p] < pu.upper_bet:
                pu.num_needed_to_action   += 1
                pu.is_needed_to_action[p]  = True
            p = (p + 1) % pu.num_players

    def action_raise(self, action):
        """

        Args:
            action:
        """
        pu = self.public_state

        pu.chips[pu.turn] -= action.price
        pu.bets[pu.turn]  += action.price
        pu.max_bet_sofar   = pu.bets[pu.turn]

        pu.is_needed_to_action[pu.turn] = False
        pu.num_needed_to_action        -= 1
        pu.is_raise[pu.turn]            = True
        pu.num_raise                   +=1
        p = (pu.turn + 1)%pu.num_players
        while p != pu.turn:
            if pu.is_quit[p] == False and pu.is_needed_to_action[p] == False and pu.bets[p] < pu.upper_bet:
                pu.num_needed_to_action   += 1
                pu.is_needed_to_action[p]  = True
            p = (p + 1) % pu.num_players


    def action_showhand(self, action):
        """

        Args:
            action:
        """
        pu = self.public_state

        pu.bets[pu.turn]               += action.price
        pu.chips[pu.turn]               = 0

        pu.is_needed_to_action[pu.turn] = False
        pu.num_needed_to_action        -= 1
        if pu.bets[pu.turn] > pu.max_bet_sofar:
            p = (pu.turn + 1) % pu.num_players
            while p != pu.turn:
                if pu.is_quit[p] == False and pu.is_needed_to_action[p] == False and pu.bets[p] < pu.upper_bet:
                    pu.num_needed_to_action  += 1
                    pu.is_needed_to_action[p] = True
                p = (p + 1) % pu.num_players

            pu.is_raise[pu.turn] = True
            pu.max_bet_sofar     = pu.bets[pu.turn]
            pu.num_raise         = False

############################################# Utils Function ######################################################
    @classmethod
    def check_initialization_configuratioin(cls, env):
        """

        Args:
            env:

        Returns:

        """
        if len(env.chips) != env.num_players:
            raise ValueError("len(env.chips)%d != env.num_players%d"%(len(env.chips), env.num_players))

        if env.num_players * 5 > 52:
            raise ValueError("env.num_players * 5 must be less than 51, now env.num_players = %d"%(env.num_players))

        return True

    @classmethod
    def is_compute_scores(cls, public_state):
        """

        Args:
            public_state:

        Returns:

        """
        pu = public_state

        if pu.num_quit == pu.num_players - 1:
            return True

        if pu.round == 4 and pu.num_needed_to_action == 0:
            return True
        if pu.num_needed_to_action == 0 and pu.max_bet_sofar == pu.upper_bet:
            return True

        return False

    @classmethod
    def compute_scores(cls, public_state):
        """

        Args:
            public_state:

        Returns:

        """

        if public_state.num_quit + 1 == public_state.num_players:
            player_id = 0
            for i in range(public_state.num_players):
                if public_state.is_quit[i] == False:
                    player_id = i
                    scores = [0 for k in range(public_state.num_players)]
                    for p in xrange(public_state.num_players):
                        if p == player_id:
                            scores[p] = sum(public_state.bets) - public_state.bets[p]
                        else:
                            scores[p] = -public_state.bets[p]
                    for p in xrange(public_state.num_players):
                        scores[p] /= public_state.floor_bet * 1.0
                    return scores

            raise ValueError("compute_scores error, is_quit = ", public_state.is_quit, "num_quit=", public_state.num_quit)


        max_cards = [public_state.first_hand_cards[0],\
                     public_state.second_hand_cards[0], public_state.third_hand_cards[0],\
                     public_state.fourth_hand_cards[0], public_state.fifth_hand_cards[0]]
        max_id    = 0
        for i in xrange(1, public_state.num_players):
            tmp = [public_state.first_hand_cards[i],\
                   public_state.second_hand_cards[i], public_state.third_hand_cards[i],\
                   public_state.fourth_hand_cards[i], public_state.fifth_hand_cards[i]]
            if FiveCardStudEnv.compare_cards(max_cards, tmp) < 0:
                max_cards = tmp
                max_id    = i
     
        scores = [0 for i in range(public_state.num_players)]
        for i in xrange(public_state.num_players):
            if i == max_id:
                scores[i] = sum(public_state.bets) - public_state.bets[i]
            else:
                scores[i] = -public_state.bets[i]
        for i in xrange(public_state.num_players):
            scores[i]    /= public_state.floor_bet * 1.0

        return scores

    @classmethod
    def choose_player_at_begining_of_round(cls, public_state):
        """

        Args:
            public_state:

        Returns:

        """

        round = public_state.round
        if round in [1,2,3]:
            public_cards = None
            if   round == 1: public_cards = public_state.second_hand_cards
            elif round == 2: public_cards = public_state.third_hand_cards
            elif round == 3: public_cards = public_state.fourth_hand_cards

            max_id = 0
            for i in range(public_state.num_players):
                if public_state.is_quit[i] == False:
                    max_id = i
                    break
            max_card = public_cards[max_id]

            for i in range(1, public_state.num_players):
                if FiveCardStudPokerCard.compare(max_card, public_cards[i]) < 0 and public_state.is_quit[i] == False:
                    max_card = public_cards[i]
                    max_id   = i
            return max_id

        elif round == 4:
            max_cards = [public_state.second_hand_cards[0], public_state.third_hand_cards[0],\
                         public_state.fourth_hand_cards[0], public_state.fifth_hand_cards[0]]
            max_id    = 0
            for i in xrange(1, public_state.num_players):
                tmp = [public_state.second_hand_cards[i], public_state.third_hand_cards[i], \
                       public_state.fourth_hand_cards[i], public_state.fifth_hand_cards[i]]
                if FiveCardStudEnv.compare_cards(max_cards, tmp) < 0:
                    max_cards = tmp
                    max_id    = i
            return max_id

        else:
            raise ValueError("pulic_state.round(%d) not in [1,2,3,4]"%(public_state.turn))

    @classmethod
    def next_player(self, pu):
        """

        Args:
            pu:

        Returns:

        """
        i = pu.turn
        if pu.num_needed_to_action == 0:
            return -1

        p = (i+1)%pu.num_players
        while pu.is_needed_to_action[p] == False:
            p = (p+1)%pu.num_players
        return p

    @classmethod
    def is_action_valid(cls, action, public_state, person_state):
        """

        Args:
            action:
            public_state:
            person_state:

        Returns:

        """
        if action.key not in person_state.available_actions:
            return False
        return True

    @classmethod
    def available_actions(cls, public_state, person_state):
        """

        Args:
            public_state:
            person_state:

        Returns:

        """
        pu              = public_state
        round           = pu.round
        turn            = pu.turn
        Showhand_count  = pu.upper_bet - pu.bets[turn]
        Call_count      = pu.max_bet_sofar - pu.bets[turn]

        '''
        print ""
        if pu.previous_action is not None:
            print ("previous_action", pu.previous_action.key)
        print ("turn=",turn, "round=", pu.round,"pu.bets[turn]",pu.bets[turn], "pu.bets", pu.bets,"pu.upper_bet",pu.upper_bet,"pu.max_bet_sofar",pu.max_bet_sofar)
        print ("is_quit",pu.is_quit)
        print ("is_need_action", pu.is_needed_to_action)
        print ("Call_count = ", Call_count)
        print ""
        sys.stdout.flush()
        '''

        actions = dict()
        if round  == 1 or round == 2 or round == 3:
            if pu.previous_round is None or pu.previous_round == round -1:
                ## bet
                for i in xrange(int(Call_count+1), int(pu.upper_bet-pu.bets[turn])):
                    actions["Bet_%d"%i]                 = FiveCardStudAction.lookup("Bet_%d"%i)
                ## fold
                actions["Fold_0"]                       = FiveCardStudAction.lookup("Fold_0")
                ## showhand
                actions["Showhand_%d"%(Showhand_count)] = FiveCardStudAction.lookup("Showhand_%d"%Showhand_count)
                ## Check_
                actions["Check_0"]                      = FiveCardStudAction.lookup("Check_0")
            else:
                ## fold
                actions["Fold_0"]                       = FiveCardStudAction.lookup("Fold_0")
                ## showhand
                actions["Showhand_%d"%Showhand_count]   = FiveCardStudAction.lookup("Showhand_%d"%(Showhand_count))
                ## Call
                if Call_count  < Showhand_count:
                    if Call_count == 0:
                        actions["Check_0"]                 = FiveCardStudAction.lookup("Check_0")
                    else:
                        actions["Call_%d"%(Call_count )]   = FiveCardStudAction.lookup("Call_%d"%(Call_count))
                ## "raise"
                if pu.is_raise[turn] == False:
                    for i in xrange(int(Call_count + 1),int(Showhand_count)):
                        actions["Raise_%d"%(i)] = FiveCardStudAction.lookup("Raise_%d"%i)


        elif round == 4:
            if pu.previous_round == round - 1:
                ## showhand
                actions["Showhand_%d"%(Showhand_count)] = FiveCardStudAction.lookup("Showhand_%d"%(Showhand_count))
                ## bet
                for i in xrange( Call_count + 1, int(pu.upper_bet) - int(pu.bets[turn])):
                    actions["Bet_%d"%i] = FiveCardStudAction.lookup("Bet_%d"%i)
                ## fold
                actions["Fold_0"]     = FiveCardStudAction.lookup("Fold_0")

            else:
                ## fold
                actions["Fold_0"]     = FiveCardStudAction("Fold_0")
                ## Call
                if Call_count  == Showhand_count:
                    actions["Showhand_%d"%(Call_count)]       = FiveCardStudAction.lookup("Showhand_%d"%(Call_count))
                elif Call_count == 0:
                    actions["Check_0"]                        = FiveCardStudAction.lookup("Check_0")
                else:
                    actions["Call_%d"%(Call_count )]          = FiveCardStudAction.lookup("Call_%d"%(Call_count))

        else:
            raise ValueError("pulic_state.round(%d) not in [1,2,3,4]" % (public_state.turn))

        return actions

    @classmethod
    def is_nextround(self, public_state):
        '''
        :return: 
        A boolean variable indicates whether is it time to enter the next stage
        '''
        return public_state.num_needed_to_action == 0

    @classmethod
    def compare_cards(cls, cards1, cards2):
        """

        Args:
            cards1:
            cards2:

        Returns:

        """
        if len(cards1) == len(cards2) and len(cards1) == 4:
            pattern1 = cls.fourcards2pattern(cards1)
            pattern2 = cls.fourcards2pattern(cards2)
            if pattern1[5] != pattern2[5]:
                return pattern1[5] - pattern2[5]
            else:
                cards1.sort(FiveCardStudPokerCard.compare)
                cards2.sort(FiveCardStudPokerCard.compare)
                return FiveCardStudPokerCard.compare(cards1[-1], cards2[-1])

        elif len(cards1) == len(cards2) and len(cards1) == 5:
            pattern1 = cls.cards2pattern(cards1)
            pattern2 = cls.cards2pattern(cards2)
            if pattern1[5] != pattern2[5]:
                return pattern1[5] - pattern2[5]
            else:
                cards1.sort(FiveCardStudPokerCard.compare)
                cards2.sort(FiveCardStudPokerCard.compare)
                return FiveCardStudPokerCard.compare(cards1[-1], cards2[-1])

        else:
            raise  ValueError("len(cards1)%d, and len(cards2)%d are same and are 4 or 5 "%(len(cards1),len(cards2)))

    @classmethod
    def cards2pattern(cls, cards):
        """

        Args:
            cards:

        Returns:

        """
        pointrank2cards = dict()
        for c in cards:
            if c.get_point_rank() in pointrank2cards:
                pointrank2cards[c.get_point_rank()].append(c)
            else:
                pointrank2cards[c.get_point_rank()] = [c]
        for p in pointrank2cards:
            pointrank2cards[p].sort(FiveCardStudPokerCard.compare)

        suitrank2cards = dict()
        for c in cards:
            if c.get_suit_rank() in suitrank2cards:
                suitrank2cards[c.get_suit_rank()].append(c)
            else:
                suitrank2cards[c.get_suit_rank()] = [c]
        for s in suitrank2cards:
            suitrank2cards[s].sort(FiveCardStudPokerCard.compare)

        num2pointrank = [[], [], [], [], []]
        for p in pointrank2cards:
            num = len(pointrank2cards[p])
            num2pointrank[num].append(p)
        for i in xrange(5):
            num2pointrank[num].sort()

        sorted_pointrank = []
        for p in pointrank2cards:
            sorted_pointrank.append(p)
        sorted_pointrank.sort()

        ##straight_samesuit
        for s in suitrank2cards:
            if len(suitrank2cards[s]) >= 5:
                numStraight = 1
                for i in xrange(len(suitrank2cards[s]) - 2, -1, -1):
                    if suitrank2cards[s][i].get_point_rank() == suitrank2cards[s][i + 1].get_point_rank() - 1:
                        numStraight += 1
                    else:
                        numStraight = 1

                    if numStraight == 5:
                        pattern = roomai.fivecardstud.FiveCardStudAllCardsPattern["Straight_SameSuit"]
                        return pattern

        ##4_1
        if len(num2pointrank[4]) ==1:
            pattern =  roomai.fivecardstud.FiveCardStudAllCardsPattern["4_1"]
            return pattern

        ##3_2
        if len(num2pointrank[3]) == 1 and len(num2pointrank[2]) == 1:
            pattern =  roomai.fivecardstud.FiveCardStudAllCardsPattern["3_2"]
            return pattern

        ##SameSuit
        for s in suitrank2cards:
            if len(suitrank2cards[s]) >= 5:
                pattern =  roomai.fivecardstud.FiveCardStudAllCardsPattern["SameSuit"]
                return pattern

        ##Straight_DiffSuit
        numStraight = 1
        for idx in xrange(len(sorted_pointrank) - 2, -1, -1):
            if sorted_pointrank[idx] + 1 == sorted_pointrank[idx]:
                numStraight += 1
            else:
                numStraight = 1

            if numStraight == 5:
                pattern =  roomai.fivecardstud.FiveCardStudAllCardsPattern["Straight_DiffSuit"]
                for p in xrange(idx, idx + 5):
                    point = sorted_pointrank[p]
                return pattern

        ##3_1_1
        if len(num2pointrank[3]) == 1:
            pattern =  roomai.fivecardstud.FiveCardStudAllCardsPattern["3_1_1"]
            return pattern

        ##2_2_1
        if len(num2pointrank[2]) >= 2:
            pattern =  roomai.fivecardstud.FiveCardStudAllCardsPattern["2_2_1"]
            return pattern

        ##2_1_1_1
        if len(num2pointrank[2]) == 1:
            pattern =  roomai.fivecardstud.FiveCardStudAllCardsPattern["2_1_1_1"]
            return pattern

        ##1_1_1_1_1
        return  roomai.fivecardstud.FiveCardStudAllCardsPattern["1_1_1_1_1"]

    @classmethod
    def fourcards2pattern(cls, cards):
        """

        Args:
            cards:

        Returns:

        """
        pointrank2cards = dict()
        for c in cards:
            if c.get_point_rank() in pointrank2cards:
                pointrank2cards[c.get_point_rank()].append(c)
            else:
                pointrank2cards[c.get_point_rank()] = [c]
        for p in pointrank2cards:
            pointrank2cards[p].sort(FiveCardStudPokerCard.compare)

        suitrank2cards = dict()
        for c in cards:
            if c.get_suit_rank() in suitrank2cards:
                suitrank2cards[c.get_suit_rank()].append(c)
            else:
                suitrank2cards[c.get_suit_rank()] = [c]
        for s in suitrank2cards:
            suitrank2cards[s].sort(FiveCardStudPokerCard.compare)

        num2pointrank = [[], [], [], [], []]
        for p in pointrank2cards:
            num = len(pointrank2cards[p])
            num2pointrank[num].append(p)
        for i in xrange(5):
            num2pointrank[num].sort()

        sorted_pointrank = []
        for p in pointrank2cards:
            sorted_pointrank.append(p)
        sorted_pointrank.sort()

        ##candidate straight_samesuit
        for s in suitrank2cards:
            if len(suitrank2cards[s]) >= 4:
                numStraight = 1
                for i in xrange(len(suitrank2cards[s]) - 2, -1, -1):
                    if suitrank2cards[s][i].get_point_rank() == suitrank2cards[s][i + 1].get_point_rank()  - 1:
                        numStraight += 1
                    else:
                        numStraight = 1

                    if numStraight == 4:
                        pattern = roomai.fivecardstud.FiveCardStudAllCardsPattern["Straight_SameSuit"]
                        return pattern

        ##4_1
        if len(num2pointrank[4]) == 1:
            pattern = roomai.fivecardstud.FiveCardStudAllCardsPattern["4_1"]
            return pattern

        ##3_2 impossible
        if len(num2pointrank[3]) == 1 and len(num2pointrank[2]) == 1:
            pattern = roomai.fivecardstud.FiveCardStudAllCardsPattern["3_2"]
            return pattern

        ##SameSuit
        for s in suitrank2cards:
            if len(suitrank2cards[s]) >= 4:
                pattern = roomai.fivecardstud.FiveCardStudAllCardsPattern["SameSuit"]
                return pattern

        ##Straight_DiffSuit
        numStraight = 1
        for idx in xrange(len(sorted_pointrank) - 2, -1, -1):
            if sorted_pointrank[idx] + 1 == sorted_pointrank[idx]:
                numStraight += 1
            else:
                numStraight = 1

            if numStraight == 4:
                pattern = roomai.fivecardstud.FiveCardStudAllCardsPattern["Straight_DiffSuit"]
                return pattern

        ##3_1_1
        if len(num2pointrank[3]) == 1:
            pattern = roomai.fivecardstud.FiveCardStudAllCardsPattern["3_1_1"]
            return pattern

        ##2_2_1
        if len(num2pointrank[2]) >= 2:
            pattern = roomai.fivecardstud.FiveCardStudAllCardsPattern["2_2_1"]
            return pattern

        ##2_1_1_1
        if len(num2pointrank[2]) == 1:
            pattern = roomai.fivecardstud.FiveCardStudAllCardsPattern["2_1_1_1"]
            return pattern

        ##1_1_1_1_1
        return roomai.fivecardstud.FiveCardStudAllCardsPattern["1_1_1_1_1"]



#!/bin/python
#coding:utf-8
import random
import roomai.common
import roomai.kuhn.KuhnPokerUtils

class KuhnPokerAlwaysBetPlayer(roomai.common.AbstractPlayer):
    """
    """
    def __init__(self):
        """

        """
        pass
           
    def receive_info(self, info):
        """

        Args:
            info:
        """
        pass     

    def take_action(self):
        """

        Returns:

        """
        return roomai.kuhn.KuhnPokerUtils.KuhnPokerAction("bet")

    def reset(self):
        """

        """
        pass


class KuhnPokerRandomPlayer(roomai.common.AbstractPlayer):
    """
    """
    #@override
    def receive_info(self, info):
        """

        Args:
            info:
        """
        if info.person_state.available_actions is not None:
            self.available_actions = info.person_state.available_actions

    #@override
    def take_action(self):
        """

        Returns:

        """
        idx = int(random.random() * len(self.available_actions))
        return self.available_actions.values()[idx]

    #@overide
    def reset(self):
        """

        """
        pass



#!/bin/python
#coding:utf-8

import os
import roomai.common
import copy
import itertools



'''
class Utils:
    gen_allactions = False



    @classmethod
    def lookup_action(cls, masterCards, slaveCards):
        masterCards.sort()
        slaveCards.sort()

        key_int = (masterCards + slaveCards)
        key_str = []
        for i in key_int:
            key_str.append(DouDiZhuActionElement.rank_to_str[i])
        key_str.sort()
        key = "".join(key_str)

        if cls.gen_allactions == True:
            return key, DouDiZhuPokerAction(masterCards, slaveCards)

        if key in AllActions:
            return key, AllActions[key]
        else:
            raise Exception(key + "is not in AllActions")

    @classmethod
    def lookup_action_by_key(cls, key):
        if key in AllActions:
            return key, AllActions[key]
        else:
            raise Exception(key + "is not in AllActions")

    @classmethod
    def candidate_actions(cls, hand_cards, public_state):

        patterns = []
        if public_state.phase == PhaseSpace.bid:
            patterns.append(AllPatterns["i_cheat"])
            patterns.append(AllPatterns["i_bid"])
        else:
            if public_state.is_response == False:
                for p in AllPatterns:
                    if p != "i_cheat" and p != "i_invalid":
                        patterns.append(AllPatterns[p])
            else:
                patterns.append(public_state.license_action.pattern)
                if public_state.license_action.pattern[6] == 1:
                    patterns.append(AllPatterns["p_4_1_0_0_0"])  # rank = 10
                    patterns.append(AllPatterns["x_rocket"])  # rank = 100
                if public_state.license_action.pattern[6] == 10:
                    patterns.append(AllPatterns["x_rocket"])  # rank = 100
                patterns.append(AllPatterns["i_cheat"])

        is_response = public_state.is_response
        license_act = public_state.license_action
        actions = dict()

        for pattern in patterns:
            numMaster = pattern[1]
            numMasterPoint = pattern[2]
            isStraight = pattern[3]
            numSlave = pattern[4]
            MasterCount = -1
            SlaveCount = -1

            if numMaster > 0:
                MasterCount = int(numMaster / numMasterPoint)

            if "i_invalid" == pattern[0]:
                continue

            if "i_cheat" == pattern[0]:
                key, action = cls.lookup_action([DouDiZhuActionElement.cheat], [])
                if cls.is_action_valid(hand_cards, public_state, action) == True:
                    actions[key] = action
                continue

            if "i_bid" == pattern[0]:
                key, action = cls.lookup_action([DouDiZhuActionElement.bid], [])
                if cls.is_action_valid(hand_cards, public_state, action) == True:
                    actions[key] = action
                continue

            if pattern[0] == "x_rocket":
                if hand_cards.cards[DouDiZhuActionElement.r] == 1 and \
                                hand_cards.cards[DouDiZhuActionElement.R] == 1:
                    key, action = cls.lookup_action([DouDiZhuActionElement.r, DouDiZhuActionElement.R], [])
                    if cls.is_action_valid(hand_cards, public_state, action) == True:
                        actions[key] = action
                continue

            if pattern[1] + pattern[4] > hand_cards.num_cards:
                continue
            sum1 = 0

            for count in range(MasterCount, 5, 1):
                sum1 += hand_cards.count2num[count]
            if sum1 < numMasterPoint:
                continue

            ### action with cards
            mCardss = []
            mCardss = Utils.extractMasterCards(hand_cards, numMasterPoint, MasterCount, pattern)

            for mCards in mCardss:
                if numSlave == 0:
                    key, action = cls.lookup_action(mCards, [])
                    if cls.is_action_valid(hand_cards, public_state, action) == True:
                        actions[key] = action
                    continue

                sCardss = Utils.extractSlaveCards(hand_cards, numSlave, mCards, pattern)
                for sCards in sCardss:
                    key, action = cls.lookup_action(mCards, sCards)
                    if cls.is_action_valid(hand_cards, public_state, action) == True:
                        actions[key] = action
        return actions


'''



#!/bin/python
import roomai.common
import random

class SevenKingRandomPlayer(roomai.common.AbstractPlayer):
    """
    """
    #@override
    def receive_info(self, info):
        """

        Args:
            info:
        """
        self.available_actions = info.person_state.available_actions

    #@override
    def take_action(self):
        """

        Returns:

        """
        idx = int(random.random() * len(self.available_actions))
        return self.available_actions.values()[idx]

    #@overide
    def reset(self):
        """

        """
        pass
